"""
DataTransform MDB module - class for performing key lookup
using conversions described in a networkx graph.
"""
# pylint: disable=E0401, E0611
import copy

from networkx import all_simple_paths, all_shortest_paths, nx
from pymongo.collation import Collation
from biothings.hub.datatransform import DataTransform
from biothings.hub.datatransform import DataTransformEdge
from biothings.hub.datatransform import IDStruct
from biothings.hub.datatransform import nested_lookup
import biothings.utils.mongo as mongo
from biothings.utils.loggers import get_logger


class MongoDBEdge(DataTransformEdge):
    """
    The MongoDBEdge uses data within a MongoDB collection to convert
    one identifier to another. The input identifier is used to search
    a collection. The output identifier values are read out of that
    collection:
    """
    def __init__(self, collection_name, lookup, field, weight=1, label=None, check_index=True):
        # pylint: disable=R0913
        """
        :param collection_name: The name of the MongoDB collection.
        :type collection_name: str
        :param lookup: The field that will match the input identifier in the collection.
        :type lookup: str
        :param field: The output identifier field that will be read out of matching documents.
        :type field: str
        :param weight: Weights are used to prefer one path over another.
                       The path with the lowest weight is preferred.
                       The default weight is 1.
        :type weight: int
        """

        super(MongoDBEdge, self).__init__(label)
        # unpickleable attributes, grouped
        self.init_state()
        self.collection_name = collection_name
        self.lookup = lookup
        self.field = field
        self.weight = weight
        if check_index:
            if self.collection_name in self.collection.database.collection_names():
                avail_idxs = {}
                for idx in self.collection.list_indexes():
                    keys = idx["key"]
                    # this could be a composite index, multiple keys being part of the index
                    # we'll consider them as individually accessible, but I'm not sure how
                    # MongoDB deals with that => TODO check
                    for k in keys:
                        avail_idxs[k] = True
                if self.lookup not in avail_idxs:
                    raise ValueError("Field '%s' isn't indexed, this would " % self.lookup
                                     + "result in very long datatransform process")
            else:
                self.logger.warning("Collection '%s' doesn't exist, can't check indices" % self.collection_name)

    def init_state(self):
        self._state = {
            "collection": None,
            "logger": None
        }

    @property
    def collection(self):
        """getting for collection member variable"""
        if not self._state["collection"]:
            try:
                self.prepare_collection()
            # pylint: disable=W0703
            except Exception:
                # if accessed but not ready, then just ignore and return invalid
                # value for a client
                return None
        return self._state["collection"]

    def prepare_collection(self):
        """
        Load the mongodb collection specified by collection_name.
        :return:
        """
        self._state["collection"] = mongo.get_src_db()[self.collection_name]
        self.logger.info("Registering collection:  {}".format(self.collection_name))

    def edge_lookup(self, keylookup_obj, id_strct, debug=False):
        """
        Follow an edge given a key.

        An edge represets a document and this method uses the data in the edge_object
        to find one key to another key using exactly one mongodb lookup.
        :param keylookup_obj:
        :param id_strct:
        :return:
        """
        if not isinstance(id_strct, IDStruct):
            raise TypeError("edge_lookup id_struct is of the wrong type")

        # Build up a new_id_strct from the results
        res_id_strct = IDStruct()

        # Keep the old debug information
        if debug:
            res_id_strct.import_debug(id_strct)

        id_lst = id_strct.id_lst
        if id_lst:
            find_lst = self.collection_find(id_lst, self.lookup, self.field)

            for doc in find_lst:
                for orig_id in id_strct.find_right(nested_lookup(doc, self.lookup)):
                    res_id_strct.add(orig_id, nested_lookup(doc, self.field))
                    if debug:
                        res_id_strct.set_debug(orig_id, self.label, nested_lookup(doc, self.field))
        return res_id_strct

    def collection_find(self, id_lst, lookup, field):
        """
        Abstract out (as one line) the call to collection.find
        """
        return self.collection.find({lookup: {"$in": id_lst}}, {lookup: 1, field: 1})


class CIMongoDBEdge(MongoDBEdge):
    """
    Case-insensitive MongoDBEdge
    """
    def __init__(self, collection_name, lookup, field, weight=1, label=None):
        # pylint: disable=R0913, W0235
        super(CIMongoDBEdge, self).__init__(collection_name, lookup, field, weight, label)

    def collection_find(self, id_lst, lookup, field):
        """
        Abstract out (as one line) the call to collection.find
        and use a case-insensitive collation
        """
        return self.collection.find({lookup: {"$in": id_lst}}, {lookup: 1, field: 1})\
            .collation(Collation(locale='en', strength=2))


class DataTransformMDB(DataTransform):
    """
    Convert document identifiers from one type to another.
    """
    # Constants
    batch_size = 1000
    default_source = '_id'

    def __init__(self, graph, *args, **kwargs):
        """
        The DataTransformNetworkX module was written as a decorator class
        which should be applied to the load_data function of a
        Biothings Uploader.  The load_data function yields documents,
        which are then post processed by call and the 'id' key
        conversion is performed.

        :param graph: nx.DiGraph (networkx 2.1) configuration graph
        :param input_types: A list of input types for the form (identifier, field) where
                            identifier matches a node and field is an optional dotstring
                            field for where the identifier should be read from
                            (the default is '_id').
        :param output_types: A priority list of identifiers to convert to. These
                             identifiers should match nodes in the graph.
        :type output_types: list(str)
        :param id_priority_list: A priority list of identifiers to to sort input
                                 and output types by.
        :type id_priority_list: list(str)
        :param skip_on_failure: If True, documents where identifier conversion fails
                                will be skipped in the final document list.
        :type skip_on_failure: bool
        :param skip_w_regex: Do not perform conversion if the identifier matches
                             the regular expression provided to this argument. By default,
                             this option is disabled.
        :type skip_w_regex: bool
        :param skip_on_success: If True, documents where identifier conversion succeeds
                                will be skipped in the final document list.
        :type skip_on_success: bool
        :param idstruct_class: Override an internal data structure used by the this
                               module (advanced usage)
        :type idstruct_class: class
        :param copy_from_doc: If true then an identifier is copied from the input
                              source document regardless as to weather it matches an
                              edge or not. (advanced usage)
        :type copy_from_doc: bool
        """
        if not isinstance(graph, nx.DiGraph):
            raise ValueError("key_lookup configuration error:  graph must be of type nx.DiGraph")
        self._validate_graph(graph)
        self.graph = graph
        self.logger, _ = get_logger('datatransform')

        super(DataTransformMDB, self).__init__(*args, **kwargs)
        self._precompute_paths()

    def _valid_input_type(self, input_type):
        return input_type.lower() in self.graph.nodes()

    def _valid_output_type(self, output_type):
        return output_type.lower() in self.graph.nodes()

    @staticmethod
    def _validate_graph(graph):
        """
        Check if the input configuration graph graph has a valid structure.
        :param graph: key_lookup configuration graph
        :return:
        """
        # all node names should be lowercase
        for node in graph.nodes():
            if node != node.lower():
                raise ValueError("node object {} is not lowercase".format(node))
        for (vert1, vert2) in graph.edges():
            if 'object' not in graph.edges[vert1, vert2].keys():
                raise ValueError("edge_object for ({}, {}) is missing".format(vert1, vert2))
            edge_object = graph.edges[vert1, vert2]['object']
            if not isinstance(edge_object, DataTransformEdge):
                raise ValueError("edge_object for ({}, {}) is of the wrong type".
                                 format(vert1, vert2))

    def _precompute_paths(self):
        """
        Precompute all paths from the given key_type to all target key types
        provided on initialization.
        :return:
        """
        self.paths = {}
        for output_type in self.output_types:
            for input_type in self.input_types:
                paths = [p for p in all_simple_paths(self.graph, input_type[0], output_type)]
                if not paths:
                    try:
                        # this will try to find self-loops. all_shortest_paths() return one element,
                        # the self-lopped node, but we need an tuple so the "*2"
                        # also make sure those self-loops actually are defined in the graph
                        try:
                            # this will raise a keyerror is edge for self-loop
                            # p-to-p isn't defined
                            # pylint: disable=W0104
                            self.graph.edges[input_type[0], output_type]
                            paths = [p*2 for p in all_shortest_paths(
                                self.graph, input_type[0], output_type)]
                        except KeyError:
                            pass
                    except nx.NetworkXNoPath:
                        pass
                # Sort by path length - try the shortest paths first
                paths = sorted(paths, key=self._compute_path_weight)
                self.paths[(input_type[0], output_type)] = paths
        # self.logger.debug("All Pre-Computed DataTransform Paths:  {}".format(self.paths))

    def key_lookup_batch(self, batchiter):
        # pylint: disable=R0912
        """
        Look up all keys for ids given in the batch iterator (1 block)
        :param batchiter:  1 lock of records to look up keys for
        :return:
        """
        doc_lst = []
        for doc in batchiter:
            # in debug mode, skip all documents not in the debug list
            if self.debug:
                # pylint: disable=C0121
                if self.debug is True or doc['_id'] in self.debug:
                    # set debug information
                    doc['dt_debug'] = {'orig_id': doc['_id']}
                    doc_lst.append(doc)
            else:
                doc_lst.append(doc)

        hit_lst = []
        miss_lst = []
        for doc in doc_lst:
            if self.skip_w_regex and self.skip_w_regex.match(doc['_id']):
                yield doc
            else:
                miss_lst.append(doc)

        # Attempt to reach each destination in order...
        for output_type in self.output_types:
            # Starting with each input_type
            for input_type in self.input_types:
                # self.logger.debug("Attempt Lookup:  from '{}' To '{}'"\
                # .format(input_type[0], output_type))
                if output_type == input_type[0]:
                    # the doc itself has the correct ID,
                    # so either there's a self-loop avail to check this ID is valid
                    if self.graph.has_edge(output_type, output_type):
                        (hit_lst, miss_lst) = self.travel(input_type, output_type, miss_lst)
                    # or if copy is allowed, we get the value from the doc
                    elif self.copy_from_doc:
                        (hit_lst, miss_lst) = self._copy(input_type, miss_lst)
                else:
                    (hit_lst, miss_lst) = self.travel(input_type, output_type, miss_lst)

                if not self.skip_on_success:
                    for doc in hit_lst:
                        yield doc

        # Keep the misses if we do not skip on failure
        if not self.skip_on_failure:
            for doc in miss_lst:
                yield doc

    def _copy(self, input_type, doc_lst):
        """Copy ids in the case where input_type == output_type"""
        hit_lst = []
        miss_lst = []
        for doc in doc_lst:
            val = nested_lookup(doc, input_type[1])
            if val:
                # ensure _id is always a str
                doc['_id'] = str(val)
                hit_lst.append(doc)
                # retain debug information if available (assumed dt_debug already in place)
                if self.debug:
                    doc['dt_debug']['copy_from'] = (input_type[1], val)
            else:
                miss_lst.append(doc)
        # Keep a record of IDs copied
        self.histogram.update_io(input_type, input_type, len(hit_lst))
        return (hit_lst, miss_lst)

    def _compute_path_weight(self, path):
        """
        Helper function to compute the weight of a path
        :param path:
        :return: computed weight
        """
        weight = 0
        for path_var in map(nx.utils.pairwise, [path]):
            for (vert1, vert2) in path_var:
                edge = self.graph.edges[vert1, vert2]['object']
                weight = weight + edge.weight
        return weight

    def travel(self, input_type, target, doc_lst):
        # pylint: disable=R0914
        """
        Traverse a graph from a start key type to a target key type using
        precomputed paths.

        :param start: key type to start from
        :param target: key type to end at
        :param key: key value of type 'start'
        :return:
        """

        def _build_path_strct(input_type, doc_lst):
            """
            Build the path structure for the travel function
            :return:
            """
            return self.idstruct_class(input_type[1], doc_lst)

        def _build_hit_miss_lsts(doc_lst, id_strct, debug):
            """
            Return a list of documents that have had their identifiers replaced
            also return a list of documents that were not changed
            :param doc_lst:
            :param id_strct:
            :return:
            """
            hit_lst = []
            miss_lst = []
            for doc in doc_lst:
                hit_flag = False
                value = nested_lookup(doc, input_type[1])
                for lookup_id in id_strct.find_left(value):
                    new_doc = copy.deepcopy(doc)
                    # ensure _id is always a str
                    new_doc['_id'] = str(lookup_id)
                    # capture debug information
                    if debug:
                        new_doc['dt_debug']['start_field'] = input_type[1]
                        new_doc['dt_debug']['debug'] = id_strct.get_debug(value)
                    hit_lst.append(new_doc)
                    hit_flag = True
                if not hit_flag:
                    miss_lst.append(doc)
            return hit_lst, miss_lst

        #self.logger.debug("Travel From '{}' To '{}'".format(input_type[0], target))

        # Keep a running list of all saved hits
        saved_hits = IDStruct()

        # Build the path structure, which will save results
        path_strct = _build_path_strct(input_type, doc_lst)

        for path in map(nx.utils.misc.pairwise, self.paths[(input_type[0], target)]):
            for (vert1, vert2) in path:
                edge = self.graph.edges[vert1, vert2]['object']
                num_input_ids = len(path_strct)
                path_strct = self._edge_lookup(edge, path_strct)
                num_output_ids = len(path_strct)
                if num_input_ids:
                    # self.logger.debug("Edge {} - {}, {} searched returned {}"\
                    #        .format(vert1, vert2, num_input_ids, num_output_ids))
                    self.histogram.update_edge(vert1, vert2, num_output_ids)

            if path_strct:
                saved_hits += path_strct

            # reset the state to lookup misses
            path_strct = self.idstruct_class()
            for doc in doc_lst:
                val = nested_lookup(doc, input_type[1])
                if val:
                    if not saved_hits.left(val):
                        path_strct.add(val, val)

        # Return a list of documents that have had their identifiers replaced
        # also return a list of documents that were not changed
        hit_lst, miss_lst = _build_hit_miss_lsts(doc_lst, saved_hits, self.debug)
        self.histogram.update_io(input_type, target, len(hit_lst))
        return hit_lst, miss_lst

    def _edge_lookup(self, edge_obj, id_strct):
        """
        Follow an edge given a key.

        This method uses the data in the edge_object
        to find one key to another key using one of
        several types of lookup functions.
        """
        return edge_obj.edge_lookup(self, id_strct, self.debug)



import time
from datetime import datetime
import asyncio
from functools import partial
import random
import math

from biothings.utils.hub_db import get_src_dump, get_source_fullname
from biothings.utils.common import timesofar
from biothings.utils.dataload import dict_traverse
from biothings.utils.mongo import id_feeder, doc_feeder
from biothings.utils.loggers import get_logger
from biothings.hub import INSPECTOR_CATEGORY
from biothings.hub.databuild.backend import create_backend
import biothings.utils.inspect as btinspect
import biothings.utils.es as es

from biothings.utils.manager import BaseManager
from biothings import config
logging = config.logger


class InspectorError(Exception):
    pass


# commong function used to call inspector
def inspect_data(backend_provider, ids, mode, pre_mapping, **kwargs):
    col = create_backend(backend_provider).target_collection
    cur = doc_feeder(col, step=len(ids), inbatch=False, query={'_id': {'$in': ids}})
    res = btinspect.inspect_docs(cur, mode=mode, pre_mapping=pre_mapping,
                                 metadata=False, auto_convert=False, **kwargs)
    return res


class InspectorManager(BaseManager):

    def __init__(self, upload_manager, build_manager, *args, **kwargs):
        super(InspectorManager, self).__init__(*args, **kwargs)
        self.upload_manager = upload_manager
        self.build_manager = build_manager
        self.logfile = None
        self.setup_log()

    def clean_stale_status(self):
        src_dump = get_src_dump()
        srcs = src_dump.find()
        for src in srcs:
            dirty = False
            jobs = src.get("inspect", {}).get("jobs", {})
            for subsrc in jobs:
                if jobs[subsrc].get("status") == "inspecting":
                    logging.warning("Found stale datasource '%s', marking inspect status as 'canceled'" % src["_id"])
                    jobs[subsrc]["status"] = "canceled"
                    dirty = True
            if dirty:
                src_dump.replace_one({"_id": src["_id"]}, src)

    def setup_log(self):
        """Setup and return a logger instance"""
        self.logger, self.logfile = get_logger('inspect')

    def inspect(self, data_provider, mode="type", batch_size=10000,
                limit=None, sample=None, **kwargs):
        """
        Inspect given data provider:
        - backend definition, see bt.hub.dababuild.create_backend for
          supported format), eg "merged_collection" or ("src","clinvar")
        - or callable yielding documents
        Mode:
        - "type": will inspect and report type map found in data (internal/non-standard format)
        - "mapping": will inspect and return a map compatible for later
          ElasticSearch mapping generation (see bt.utils.es.generate_es_mapping)
        - "stats": will inspect and report types + different counts found in
          data, giving a detailed overview of the volumetry of each fields and sub-fields
        - "jsonschema", same as "type" but result is formatted as json-schema standard
        - limit: when set to an integer, will inspect only x documents.
        - sample: combined with limit, for each document, if random.random() <= sample (float),
          the document is inspected. This option allows to inspect only a sample of data.
        """
        # /!\ attention: this piece of code is critical and not easy to understand...
        # Depending on the source of data to inspect, this method will create an
        # uploader or a builder. These objects don't be behave the same while they
        # pass through pickle: uploader needs to be "unprepare()"ed so it can be
        # pickled (remove some db connection, socket), while builder must *not* be
        # unprepare() because it would reset the underlying target_name (the actual
        # target collection). Also, the way results and statuses are registered is
        # different for uploader and builder...
        # So, there are lots of "if", be careful if you want to modify that code.

        data_provider_type = None  # where to register results (if possible to do so)
        registerer_obj = None  # who should register result
        t0 = time.time()
        started_at = datetime.now()
        self.logger.info("Inspecting data with mode %s and data_provider %s" % (repr(mode), repr(data_provider)))
        if callable(data_provider):
            raise NotImplementedError("data_provider as callable untested...")
        else:
            if data_provider[0] == "src":
                data_provider_type = "source"
                # find src_dump doc
                # is it a full source name (dot notation) ?
                fullname = get_source_fullname(data_provider[1])
                if fullname:
                    # it's a dot-notation
                    src_name = fullname.split(".")[0]
                else:
                    # no subsource, full source name is the passed name
                    src_name = data_provider[1]
                    fullname = src_name
                doc = get_src_dump().find_one({"_id": src_name})  # query by main source
                if not doc:
                    raise InspectorError("Can't find document associated to '%s'" % src_name)
                # get an uploader instance (used to get the data if type is "uploader"
                # but also used to update status of the datasource via register_status()
                ups = self.upload_manager[fullname]  # potentially using dot notation
                if len(ups) > 1:
                    # recursively call inspect(), collect and return corresponding tasks
                    self.logger.debug("Multiple uploaders found, running inspector for each of them: %s" % ups)
                    res = []
                    for up in ups:
                        r = self.inspect((data_provider[0], "%s" % up.name), mode=mode, batch_size=batch_size,
                                         limit=limit, sample=sample, **kwargs)
                        res.append(r)
                    return res

                assert len(ups) == 1, "More than one uploader found for '%s', not supported (yet), use main_source.source notation" % data_provider[1]
                # create uploader
                registerer_obj = self.upload_manager.create_instance(ups[0])
                backend_provider = data_provider
            else:
                try:
                    data_provider_type = "build"
                    registerer_obj = self.build_manager.get_builder(data_provider)
                    backend_provider = data_provider
                except Exception as e:
                    raise InspectorError("Unable to create backend from '%s': %s" % (repr(data_provider), e))

        got_error = None
        try:
            @asyncio.coroutine
            def do():
                yield from asyncio.sleep(0.0)
                nonlocal mode

                pinfo = {"category": INSPECTOR_CATEGORY,
                         "source": "%s" % repr(data_provider),
                         "step": "",
                         "description": ""}
                # register begin of inspection (differ slightly depending on type)
                if data_provider_type == "source":
                    registerer_obj.register_status("inspecting", subkey="inspect")
                elif data_provider_type == "build":
                    registerer_obj.register_status("inspecting", transient=True, init=True, job={"step": "inspect"})

                self.logger.info("Running inspector on %s (type:%s,data_provider:%s)" %
                                 (repr(data_provider), data_provider_type, backend_provider))
                if sample is not None:
                    self.logger.info("Sample set to %s, inspect only a subset of data", sample)
                if limit is None:
                    self.logger.info("Inspecting all the documents")
                else:
                    nonlocal batch_size
                    # adjust batch_size so we inspect only "limit" docs if batch is smaller than the limit
                    if batch_size > limit:
                        batch_size = limit
                    self.logger.info("Inspecting only %s documents", limit)
                # make it pickleable
                if data_provider_type == "source":
                    # because register_obj is also used to fetch data, it has to be unprepare() for pickling
                    registerer_obj.unprepare()
                else:
                    # NOTE: do not unprepare() the builder, we'll loose the target name
                    # (it's be randomly generated again) and we won't be able to register results
                    pass

                cnt = 0
                doccnt = 0
                jobs = []
                # normalize mode param and prepare global results
                if type(mode) == str:
                    mode = [mode]

                converters, mode = btinspect.get_converters(mode)

                inspected = {}
                for m in mode:
                    inspected.setdefault(m, {})

                backend = create_backend(backend_provider).target_collection
                for ids in id_feeder(backend, batch_size=batch_size):
                    if sample is not None:
                        if random.random() > sample:
                            continue
                    cnt += 1
                    doccnt += batch_size
                    if limit and doccnt > limit:
                        break
                    pinfo["description"] = "batch #%s" % cnt

                    def batch_inspected(bnum, i, f):
                        nonlocal inspected
                        nonlocal got_error
                        nonlocal mode
                        try:
                            res = f.result()
                            for m in mode:
                                inspected[m] = btinspect.merge_record(inspected[m], res[m], m)
                        except Exception as e:
                            got_error = e
                            self.logger.error("Error while inspecting data from batch #%s: %s" % (bnum, e))
                            raise

                    pre_mapping = "mapping" in mode  # we want to generate intermediate mapping so we can merge
                    # all maps later and then generate the ES mapping from there
                    self.logger.info("Creating inspect worker for batch #%s" % cnt)
                    job = yield from self.job_manager.defer_to_process(pinfo,
                                                                       partial(inspect_data, backend_provider,
                                                                               ids, mode=mode, pre_mapping=pre_mapping, **kwargs))
                    job.add_done_callback(partial(batch_inspected, cnt, ids))
                    jobs.append(job)

                yield from asyncio.gather(*jobs)

                # compute metadata (they were skipped before)
                for m in mode:
                    if m == "mapping":
                        try:
                            inspected["mapping"] = es.generate_es_mapping(inspected["mapping"])
                            # metadata for mapping only once generated
                            inspected = btinspect.compute_metadata(inspected, m)
                        except es.MappingError as e:
                            inspected["mapping"] = {"pre-mapping": inspected["mapping"], "errors": e.args[1]}
                    else:
                        inspected = btinspect.compute_metadata(inspected, m)

                # just potential converters
                btinspect.run_converters(inspected, converters)

                def fully_inspected(res):
                    nonlocal got_error
                    try:
                        res = btinspect.stringify_inspect_doc(res)
                        _map = {"results": res}
                        _map["data_provider"] = repr(data_provider)
                        _map["started_at"] = started_at
                        _map["duration"] = timesofar(t0)

                        # when inspecting with "stats" mode, we can get huge number but mongo
                        # can't store more than 2^64, make sure to get rid of big nums there
                        def clean_big_nums(k, v):
                            # TODO: same with float/double? seems mongo handles more there ?
                            if isinstance(v, int) and v > 2**64:
                                return k, math.nan
                            else:
                                return k, v

                        dict_traverse(_map, clean_big_nums)
                        # register begin of inspection (differ slightly depending on type)
                        if "mapping" in mode and "errors" in res["mapping"] and "pre-mapping" in res["mapping"]:
                            registerer_obj.register_status("failed", subkey="inspect", inspect=_map)
                            got_error = InspectorError(res["mapping"]["errors"])
                        else:
                            if data_provider_type == "source":
                                registerer_obj.register_status("success", subkey="inspect", inspect=_map)
                            elif data_provider_type == "build":
                                registerer_obj.register_status("success", job={"step": "inspect"},
                                                               build={"inspect": _map})
                    except Exception as e:
                        self.logger.exception("Error while inspecting data: %s" % e)
                        got_error = e
                        if data_provider_type == "source":
                            registerer_obj.register_status("failed", subkey="inspect", err=repr(e))
                        elif data_provider_type == "build":
                            registerer_obj.register_status("failed", job={"err": repr(e)})
                fully_inspected(inspected)
                if data_provider_type is None:
                    return
                if got_error:
                    raise got_error
            task = asyncio.ensure_future(do())
            return task
        except Exception as e:
            self.logger.error("Error while inspecting '%s': %s" % (repr(data_provider), e))
            raise



import copy
import json
import re
from collections import defaultdict
from functools import partial

from biothings.utils.common import dotdict, split_ids


class OptionArgError(Exception):

    def __init__(self, **info):
        self.info = info

class StringTranslate():
    """
    Translate the input parameter to the desired type.
    If jsoninput is set to true, try to load str in json.
    """

    def __init__(self, jsoninput=False):

        if isinstance(jsoninput, str):
            self.jsoninput = self.str_to_bool(jsoninput)
        else:
            self.jsoninput = bool(jsoninput)

    def convert(self, value, to_type):
        if isinstance(value, to_type):
            return value
        if isinstance(value, str) and to_type == bool:  # TODO use dispatch syntax
            return self.str_to_bool(value)
        elif isinstance(value, str) and to_type == list:
            return self.str_to_list(value)
        elif isinstance(value, (int, float)):
            return self.convert(str(value), to_type)
        else:
            return self.obj_to_type(value, to_type)

    @staticmethod
    def str_to_bool(val):
        return val.lower() in ('1', 'true', 'y', 't', '')

    def str_to_list(self, val, param=''):
        if self.jsoninput:
            try:
                val = json.loads(val)
            except Exception:
                pass
        if not isinstance(val, list):
            try:
                val = split_ids(str(val))
            except ValueError as e:
                raise OptionArgError(reason=str(e), param=param)
        return val

    @staticmethod
    def obj_to_type(val, type_, param=''):
        try:
            result = (type_)(val)
        except ValueError:
            raise OptionArgError(reason=f"Expect type {type_.__name__}.", param=param)
        return result

class OptionArg():
    """
    Interpret a setting file for a keyword like
    {
        'type': list,
        'default': None,
        'max': 1000,
        'alias': [
            'fields', 'field', 'filter'
        ]
        'translations': [
           (re.compile(r'rs[0-9]+', re.I), 'dbsnp.rsid')
        ]
    }
    And apply the setting for a keyword in input args.
    """

    def __init__(self, keyword, setting):

        self.keyword = keyword
        self.setting = setting

        self.list_size_cap = 1000

    def parse(self, args, path_args, path_kwargs):

        value = args.get(self.keyword)
        jsoninput = args.get('jsoninput')  # deprecated

        if 'alias' in self.setting and not value:
            value = self._alias(args)

        if 'path' in self.setting:
            value = self._path(path_args, path_kwargs)

        if value is None:
            return self._default()

        if 'type' in self.setting:
            value = self._typify(value, jsoninput)

        if 'translations' in self.setting:
            value = self._translate(value)

        if 'enum' in self.setting:
            value = self._enum(value)

        if 'max' in self.setting or self.list_size_cap:
            value = self._max(value)

        return value

    def _path(self, path_args, path_kwargs):
        # find path args from index number
        # find path kwargs from handler regex captures
        path = self.setting['path']
        if isinstance(path, int):
            if len(path_args) <= path:
                raise OptionArgError(missing=self.keyword)
            return path_args[path]
        elif isinstance(path, str):
            if path not in path_kwargs:
                raise OptionArgError(missing=self.keyword)
            return path_kwargs[path]

    def _alias(self, args):
        aliases = self.setting['alias']
        if not isinstance(aliases, list):
            aliases = [aliases]
        for _alias in aliases:
            if _alias in args:
                return args[_alias]

    def _default(self):
        # fallback to default values or raise error
        if self.setting.get('required'):
            raise OptionArgError(missing=self.keyword)
        return self.setting.get('default')

    def _typify(self, value, jsoninput):
        # convert to the desired value type and format
        if isinstance(value, self.setting['type']):
            return value
        param = StringTranslate(jsoninput)
        return param.convert(value, self.setting['type'])

    def _translate(self, obj):
        translations = self.setting['translations']
        if isinstance(obj, str):
            for (regex, translation) in translations:
                obj = re.sub(regex, translation, obj)
            return obj
        if isinstance(obj, list):
            return [self._translate(item)
                    for item in obj]
        raise TypeError()

    def _max(self, value):
        # list size and int value validation
        if isinstance(value, list):
            max_allowed = self.setting.get('max')
            max_allowed = max_allowed or self.list_size_cap
            if len(value) > max_allowed:
                raise OptionArgError(
                    keyword=self.keyword,
                    max=max_allowed,
                    lst=value)
        elif isinstance(value, (int, float, complex)) \
                and not isinstance(value, bool):
            if 'max' in self.setting:
                if value > self.setting['max']:
                    raise OptionArgError(
                        keyword=self.keyword,
                        max=self.setting['max'],
                        num=value)
        return value

    def _enum(self, value):

        # allow only specified values
        allowed = self.setting['enum']

        if isinstance(value, list):
            for val in value:
                self._enum(val)
        elif value not in allowed:
            raise OptionArgError(
                keyword=self.keyword,
                allowed=allowed
            )
        return value

class Options():

    def __init__(self, options, groups, methods):

        self._options = options  # example: {'dev':{'type':str }}
        self._groups = groups or ()  # example: ('es', 'transform')
        self._methods = list(method.upper() for method in methods or ())  # 'GET'

    def parse(self, method, args, path_args, path_kwargs):

        result = defaultdict(dict)
        options = {}

        rules = []  # expand * to kwarg_methods setting
        if not self._methods or method in self._methods:
            rules += list(self._options['*'].items())
        rules += list(self._options[method].items())

        # method precedence: specific > *
        for keyword, setting in rules:
            options[keyword] = setting

        # setting + inputs -> arg value
        for keyword, setting in options.items():
            arg = OptionArg(keyword, setting)
            val = arg.parse(args, path_args, path_kwargs)
            # discard no default value
            if val is not None:
                if 'group' in setting:
                    group = setting['group']
                    if isinstance(group, str):
                        result[group][keyword] = val
                    else:  # assume iterable
                        for _group in group:
                            result[_group][keyword] = val
                else:  # top level keywords
                    result[keyword] = val

        # make sure all named groups exist
        for group in self._groups:
            if group not in result:
                result[group] = {}

        return dotdict(result)

class OptionSets():

    def __init__(self):

        self.options = defaultdict(partial(defaultdict, dict))
        self.groups = defaultdict()
        self.methods = defaultdict()

    def add(self, name, optionset):
        if name:
            for method, options in optionset.items():
                self.options[name][method.upper()].update(options)

    def get(self, name):

        return Options(
            self.options[name],
            self.groups[name],
            self.methods[name]
        )

    def log(self):

        res = copy.deepcopy(self.options)
        res = self._serialize(res)
        for api, groups in self.groups.items():
            res[api]['_groups'] = groups
        for api, methods in self.methods.items():
            res[api]['_methods'] = methods
        return res

    def _serialize(self, obj):
        if isinstance(obj, dict):
            for key, val in obj.items():
                obj[key] = self._serialize(val)
            return obj
        elif isinstance(obj, (list, tuple)):
            return [self._serialize(item) for item in obj]
        elif isinstance(obj, (str, int)):
            return obj
        else:  # best effort
            return str(obj)



import inspect
import types
import logging
import time
import asyncio
from functools import partial
import socket
import contextlib

import tornado.web
from biothings.hub.api.handlers.base import GenericHandler
from biothings.utils.hub import CommandInformation, CommandError, CommandDefinition


class EndpointDefinition(dict):
    pass


def generate_endpoint_for_callable(name, command, method, force_bodyargs):
    if force_bodyargs is True:
        pass
        #assert method != "get", \
        #    "Can't have force_bodyargs=True with method '%s' for command '%s'" % (method,command)
    try:
        specs = inspect.getfullargspec(command)
    except TypeError as e:
        # TODO: generate a display handler
        raise TypeError("Can't determine arguments for command '%s': %s" %
                        (command, e))

    argfrom = 0
    if isinstance(command, types.MethodType) or \
            isinstance(command, partial) and isinstance(command.func, types.MethodType):
        # skip "self" arg
        argfrom = 1
    defaultargs = {}
    args = specs.args[argfrom:]
    # defaults are listed from the end ("n latest args")
    if specs.defaults:
        for i in range(-len(specs.defaults), 0):
            defaultargs[args[i]] = specs.defaults[i]
    # ignore "self" args, assuming it's the one used when dealing with method
    #mandatargs = set(args).difference({"self"}).difference(defaultargs) or ""
    mandatargs = force_bodyargs is False and set(args).difference(
        defaultargs) or ""
    cmdargs = "{}"
    if mandatargs:
        # this is for cmd args building before submitting to the shell
        cmdargs = "{" + ",".join(["'''%s''':%s" % (v, v)
                                  for v in mandatargs]) + "}"
        # this is for signature
        mandatargs = "," + ",".join(["%s" % v for v in mandatargs])
    # generate a wrapper over the passed command
    #print({"method":method,"args":args,"defaultargs":defaultargs,"name":name,
    #       "mandatargs":mandatargs,"cmdargs":cmdargs})
    strcode = """
@asyncio.coroutine
def %(method)s(self%(mandatargs)s):
    '''%(name)s => %(command)s'''
    cmdargs = %(cmdargs)s
    reqargs = {} # holds either body or query string args
    qkwargs = {} # holds kwargs (either from body or query string)
    for k in cmdargs:
        if cmdargs[k] is None:
            raise tornado.web.HTTPError(400,reason="Bad Request (Missing argument " + k + ")")

    if "%(method)s" != "get":
        # allow to have no body at all, defaulting to empty dict (no args)
        reqargs = tornado.escape.json_decode(self.request.body or '{}')
    elif %(force_bodyargs)s == True: # force_bodyargs
        for arg in %(args)s:
            qarg = self.get_query_argument(arg,None)
            if qarg:
                reqargs[arg] = qarg;
    else:
        # extract optional args
        for arg in self.request.arguments:
            if not arg in reqargs:
                qkwargs[arg] = self.get_argument(arg)
    #print("arguments:")
    #print(%(args)s)
    #print(%(defaultargs)s)
    #print(reqargs)
    #print(qkwargs)
    for arg in %(args)s + list(reqargs.keys()) + list(qkwargs.keys()):
        if arg in %(defaultargs)s or arg in qkwargs:
            mandatory = False
        else:
            mandatory = True
        if %(force_bodyargs)s or "%(method)s" != "get":
            try:
                if mandatory:
                    # part of signature (in URL) or body args ?
                    try:
                        cmdargs[arg] # just check key exists
                    except KeyError:
                        cmdargs[arg] = reqargs[arg]
                else:
                    # check if optional has been passed or if value is taken from default
                    # (used to display/build command line with minimal info,
                    # ie. what's been passed by user)
                    try:
                        val = reqargs[arg]
                        cmdargs[arg] = val
                    except KeyError:
                        pass
            except KeyError:
                raise tornado.web.HTTPError(400,reason="Bad Request (Missing argument " + arg + ")")
        else:
            # if not default arg and arg not passed, this will raise a 400 (by tornado)
            if mandatory:
                cmdargs[arg] # check key
            else:
                try:
                    val = qkwargs[arg]
                    cmdargs[arg] = val
                except KeyError:
                    pass
    # we don't pass though shell evaluation there
    # to prevent security issue (injection)...
    strcmd = '''%(name)s''' + "("
    strcmd += ",".join([str(k) + "=" + repr(v) for k,v in cmdargs.items()])
    strcmd += ")"
    res = command(**cmdargs)
    # ... but we register the command in the shell to track it
    cmdres = shell.register_command(strcmd,res)
    if type(cmdres) == CommandInformation:# or type(cmdres) == list and type(:
        # asyncio tasks unserializable
        # but keep original one
        cmdres = CommandInformation([(k,v) for k,v in cmdres.items() if k != 'jobs'])
    self.write(cmdres)
""" % {
        "method": method,
        "args": args,
        "defaultargs": defaultargs,
        "name": name,
        "mandatargs": mandatargs,
        "cmdargs": cmdargs,
        "command": repr(command),
        "force_bodyargs": force_bodyargs
    }
    #if name == "info" or name == "builds":
    #    print(strcode)
    return strcode, mandatargs != ""


def generate_endpoint_for_composite_command(name, command, method):
    strcode = """
@asyncio.coroutine
def %(method)s(self):
    # composite commands never take arguments
    cmdres = shell.eval('''%(name)s()''',return_cmdinfo=True)
    if type(cmdres) == CommandInformation:
        # asyncio tasks unserializable
        # but keep original one
        cmdres = CommandInformation([(k,v) for k,v in cmdres.items() if k != 'jobs'])
    self.write(cmdres)
""" % {
        "method": method,
        "name": name
    }
    return strcode


def generate_endpoint_for_display(name, command, method):
    strcode = """
@asyncio.coroutine
def %(method)s(self):
    self.write(command)
""" % {
        "method": method
    }
    return strcode


def generate_handler(shell, name, command_defs):

    if not type(command_defs) == list:
        command_defs = [command_defs]

    by_suffix = {}
    for commanddef in command_defs:
        confdict = {}
        method = commanddef["method"].lower()
        cmdname = commanddef["name"]
        try:
            # retrieve the actual command from the shell
            command = shell.commands.get(cmdname)  # first try public commands
            if command is None:
                command = shell.extra_ns[cmdname]  # then private/hidden commands
            # could be directly a callable or an encapsulating CommandDefinition
            if type(command) == CommandDefinition:
                command = command["command"]
        except KeyError as e:
            raise CommandError("Command '%s' can't be found in hub shell" % e)
        force_bodyargs = commanddef.get("force_bodyargs", False)
        suffix = commanddef.get("suffix", "")
        num_mandatory = 0
        #if callable(command):
        strcode, num_mandatory = generate_endpoint_for_callable(
            cmdname, command, method, force_bodyargs)
        #elif type(command) == CompositeCommand:
        #    strcode = generate_endpoint_for_composite_command(name,command,method)
        #else:
        #    assert method == "get", "display endpoint needs a GET method for command '%s'" % name
        #    strcode = generate_endpoint_for_display(name,command,method)
        # compile the code string and eval (link) to a globals() dict
        code = compile(strcode, "<string>", "exec")
        command_globals = {}
        endpoint_ns = {
            "command": command,
            "asyncio": asyncio,
            "shell": shell,
            "CommandInformation": CommandInformation,
            "tornado": tornado
        }
        eval(code, endpoint_ns, command_globals)
        methodfunc = command_globals[method]
        confdict[method] = methodfunc

        by_suffix.setdefault(suffix, {})
        by_suffix[suffix].setdefault(method, {})["confdict"] = confdict
        by_suffix[suffix][method]["num_mandatory"] = num_mandatory
        by_suffix[suffix][method]["strcode"] = strcode

    routes = []
    # when suffix present, it gives a new handler
    # so method don't get mixed (actually overwritten) together with non-suffixed
    for i, (suffix, by_method) in enumerate(by_suffix.items()):
        confdict = {}
        methods = []
        # merge all method into same handler
        for method, dat in by_method.items():
            confdict.update(dat["confdict"])
            methods.append(method)
        handler_class = type("%s%s_handler" % (name, suffix),
                             (GenericHandler, ), confdict)
        num_mandatory = dat["num_mandatory"]
        if suffix and not suffix.startswith("/"):
            suffix = "/" + suffix
        if num_mandatory:
            # this is for more REST alike URLs (eg. /info/clinvar == /info?src=clinvar
            url = r"/%s%s%s" % (name, r"/([\w\.-]+)?" * num_mandatory, suffix)
        else:
            url = r"/%s%s" % (name, suffix)
        routes.append((url, handler_class, {"shell": shell}))
        logging.info("route: %s %s => %s" %
                     (repr([m.upper() for m in methods]), url, handler_class))

    return routes


def create_handlers(shell, command_defs):
    routes = []
    for cmdname, config in command_defs.items():
        if type(config) == list:
            # multiple endpoints per handler
            commands = config
        else:
            # normalized as a list
            commands = [
                EndpointDefinition(name=config["name"],
                                   method=config.get("method", "GET"),
                                   force_bodyargs=config.get(
                                       "force_bodyargs", False))
            ]
        try:
            routes += generate_handler(shell, cmdname, commands)
        except TypeError as e:
            logging.exception("Can't generate handler for '%s': %s" %
                              (cmdname, e))
            continue

    return routes


def generate_api_routes(shell, commands, settings={}):
    routes = create_handlers(shell, commands)
    return routes


def start_api(app, port, check=True, wait=5, retry=5, settings={}):
    if check:
        # check if port is used
        def check_socket(host, port):
            num = 1
            with contextlib.closing(
                    socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                if sock.connect_ex((host, port)) == 0:
                    if num >= retry:
                        raise Exception("Can't start API, port %s is already used " % port
                                        + "and already tried %s times" % retry)
                    logging.info(
                        "Port %s is already used, sleep and retry (%s/%s)" %
                        (port, num, retry))
                    time.sleep(wait)
                    num += 1
                else:
                    return

        check_socket("localhost", port)
    app_server = tornado.httpserver.HTTPServer(app, **settings)
    app_server.listen(port)
    app_server.start()
    return app_server



import os
import subprocess
import shutil
from boto import connect_s3
from biothings.utils.aws import send_s3_file

from biothings.utils.hub_db import get_src_build
from biothings.utils.mongo import get_src_db, id_feeder, get_target_db, \
    get_cache_filename
from biothings.utils.common import anyfile

from biothings import config as btconfig
logging = btconfig.logger


def export_ids(col_name):
    """
    Export all _ids from collection named col_name.
    If col_name refers to a build where a cold_collection is defined,
    will also extract _ids and sort/uniq them to have the full list of _ids
    of the actual merged (cold+hot) collection
    Output file is stored in DATA_EXPORT_FOLDER/ids,
    defaulting to <DATA_ARCHIVE_ROOT>/export/ids. Output filename is
    returned as the end, if successful.
    """
    # prepare output directory
    DATA_EXPORT_FOLDER = getattr(btconfig, "DATA_EXPORT_FOLDER", None)
    if not DATA_EXPORT_FOLDER:
        DATA_EXPORT_FOLDER = os.path.join(btconfig.DATA_ARCHIVE_ROOT, "export")
    ids_export_folder = os.path.join(DATA_EXPORT_FOLDER, "ids")
    if not os.path.exists(ids_export_folder):
        logging.debug("Creating export/ids folder: %s" % ids_export_folder)
        os.makedirs(ids_export_folder)
    build = get_src_build().find_one({"_id": col_name})
    cold = None
    if build:
        col = get_target_db()[col_name]
        if build.get("build_config", {}).get("cold_collection"):
            cold_name = build["build_config"]["cold_collection"]
            cold = get_target_db()[cold_name]
            logging.info("Found a cold collection '%s' associated to '%s'" % (cold_name, col_name))
    else:
        # it's a src
        col = get_src_db()[col_name]

    # first iterate over all _ids. This will potentially update underlying _id cache it's not valid anymore,
    # so we're sure to work with latest data. If cache is valid, this will be pretty fast
    logging.info("Screening _ids in collection '%s'" % col.name)
    for _id in id_feeder(col, validate_only=True):
        pass
    # now accessing cache
    col_ids_cache = get_cache_filename(col.name)
    assert os.path.exists(col_ids_cache)
    logging.info("Now using cache file %s" % col_ids_cache)
    if cold:
        logging.info("Screening _ids in cold collection '%s'" % cold.name)
        for _id in id_feeder(cold, validate_only=True):
            pass
        # now accessing cache
        cold_ids_cache = get_cache_filename(cold.name)
        assert os.path.exists(cold_ids_cache)
        logging.info("Now using cache file %s" % cold_ids_cache)
    outfn = os.path.join(ids_export_folder, "%s_ids.xz" % col_name)
    # NOTE: can't use anyfile to open cache files and send _id through pipes
    # because it would load _id in memory (unless using hacks) so use cat (and
    # existing uncompressing ones, like gzcat/xzcat/...) to fully run the pipe
    # on the shell
    if cold:
        fout = anyfile(outfn, "wb")
        colext = os.path.splitext(col_ids_cache)[1]
        coldext = os.path.splitext(cold_ids_cache)[1]
        assert colext == coldext, "Hot and Cold _id cache are compressed differently (%s and %s), it should be the same" % (coldext, coldext)
        comp = colext.replace(".", "")
        supportedcomps = ["xz", "gz", ""]  # no compression allowed as well
        assert comp in supportedcomps, "Compression '%s' isn't supported (%s)" % (comp, supportedcomps)
        # IDs sent to pipe's input (sort) then compress it (xz)
        pcat = subprocess.Popen(["%scat" % comp, col_ids_cache, cold_ids_cache], stdout=subprocess.PIPE)
        psort = subprocess.Popen(["sort", "-u"], stdin=pcat.stdout, stdout=subprocess.PIPE, universal_newlines=True)
        pcat.stdout.close()  # will raise end of pipe error when finished
        if comp:
            pcomp = subprocess.Popen(["xz", "-c"], stdin=psort.stdout, stdout=fout)
        else:
            # just print stdin to stdout
            pcomp = subprocess.Popen(["tee"], stdin=psort.stdout, stdout=fout)
        psort.stdout.close()
        try:
            logging.info("Running pipe to compute list of unique _ids")
            (out, err) = pcomp.communicate()  # run the pipe! (blocking)
            if err:
                raise Exception(err)
        except Exception as e:
            logging.error("Error while running pipe to export _ids: %s" % e)
            # make sure to clean empty or half processed files
            try:
                os.unlink(outfn)
            finally:
                pass
            raise
    else:
        logging.info("Copying cache _id file")
        try:
            shutil.copyfile(col_ids_cache, outfn)
        except Exception as e:
            logging.error("Error while exporting _ids: %s" % e)
            # make sure to clean empty or half processed files
            try:
                os.unlink(outfn)
            finally:
                pass
            raise

    logging.info("Done exporting _ids to '%s'" % outfn)
    return outfn


def upload_ids(ids_file, redirect_from, s3_bucket, aws_key, aws_secret):
    """
    Upload file ids_file into s3_bucket and modify redirect_from key's metadata
    so redirect_from link will now point to ids_file
    redirect_from s3 key must exist.
    """
    # cache file should be named the same as target_name
    logging.info("Uploading _ids file %s to s3 (bucket: %s)" % (repr(ids_file), s3_bucket))
    s3path = os.path.basename(ids_file)
    send_s3_file(ids_file, s3path, overwrite=True,
                 s3_bucket=s3_bucket,
                 aws_key=aws_key,
                 aws_secret=aws_secret)
    # make the file public
    s3 = connect_s3(aws_key, aws_secret)
    bucket = s3.get_bucket(s3_bucket)
    s3key = bucket.get_key(s3path)
    s3key.set_acl("public-read")
    # update permissions and redirect metadata
    k = bucket.get_key(redirect_from)
    assert k, "Can't find s3 key '%s' to set redirection" % redirect_from
    k.set_redirect("/%s" % s3path)
    k.set_acl("public-read")
    logging.info("IDs file '%s' uploaded to s3, redirection set from '%s'" % (ids_file, redirect_from), extra={"notify": True})



"""
CIIDStruct - case insenstive id matching data structure
"""
# pylint: disable=E0611
from biothings.hub.datatransform import IDStruct


class CIIDStruct(IDStruct):
    """
    CIIDStruct - id structure for use with the DataTransform classes.  The basic idea
    is to provide a structure that provides a list of (original_id, current_id)
    pairs.

    This is a case-insensitive version of IDStruct.
    """

    def add(self, left, right):
        # pylint: disable=R0912
        """add a (original_id, current_id) pair to the list,
        All string values are typecast to lowercase"""
        if not left or not right:
            return  # identifiers cannot be None
        if self.lookup(left, right):
            return  # tuple already in the list
        # ensure it's hashable
        if not isinstance(left, (list, tuple)):
            left = [left]
        if not isinstance(right, (list, tuple)):
            right = [right]
        if isinstance(left, list):
            left = tuple(left)
        if isinstance(right, list):
            right = tuple(right)
        for val in left:
            # After some thought, this data structure should be case insensitive
            if isinstance(val, str):
                val = val.lower()
            if val not in self.forward.keys():
                self.forward[val] = right
            else:
                self.forward[val] = self.forward[val] + right
        for val in right:
            # After some thought, this data structure should be case insensitive
            if isinstance(val, str):
                val = val.lower()
            if val not in self.inverse.keys():
                self.inverse[val] = left
            else:
                self.inverse[val] = self.inverse[val] + left

    def find(self, where, ids):
        """Case insensitive lookup of ids"""
        if not ids:
            return
        if not isinstance(ids, (list, tuple)):
            ids = [ids]
        for key in ids:
            # This find is case insensitive
            if isinstance(key, str):
                key = key.lower()
            if key in where.keys():
                for i in where[key]:
                    yield i



import os
import shutil
import time
import json
from datetime import datetime
import asyncio
from functools import partial
import glob
import random

from biothings.utils.common import timesofar, get_timestamp, \
    dump, rmdashfr, loadobj, md5sum
from biothings.utils.mongo import id_feeder, get_target_db, get_previous_collection
from biothings.utils.hub_db import get_src_build
from biothings.utils.loggers import get_logger
from biothings.utils.diff import diff_docs_jsonpatch
from biothings.hub.databuild.backend import generate_folder
from biothings import config as btconfig
from biothings.utils.manager import BaseManager
from .backend import create_backend, merge_src_build_metadata
from biothings.utils.backend import DocMongoBackend
from biothings.utils.jsondiff import make as jsondiff
from biothings.hub import DIFFER_CATEGORY, DIFFMANAGER_CATEGORY
from biothings.hub.datarelease import set_pending_to_release_note

logging = btconfig.logger


class DifferException(Exception):
    pass


class BaseDiffer(object):

    # diff type name, identifying the diff algorithm
    # must be set in sub-class
    diff_type = None

    def __init__(self, diff_func, job_manager, log_folder):
        self.old = None
        self.new = None
        self.log_folder = log_folder
        self.job_manager = job_manager
        self.diff_func = diff_func
        self.timestamp = datetime.now()
        self.logfile = None
        self.setup_log()
        self.ti = time.time()
        self.metadata = {}  # diff metadata
        self.metadata_filename = None

    def setup_log(self):
        self.logger, self.logfile = get_logger(
            'diff_%s' % self.__class__.diff_type, self.log_folder)

    def get_predicates(self):
        return []

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": DIFFER_CATEGORY,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def register_status(self, status, transient=False, init=False, **extra):
        src_build = get_src_build()
        job_info = {
            'status': status,
            'step_started_at': datetime.now(),
            'logfile': self.logfile,
        }
        diff_key = self.old.target_name
        diff_info = {"diff": {diff_key: {}}}
        if transient:
            # record some "in-progress" information
            job_info['pid'] = os.getpid()
        else:
            # only register time when it's a final state
            job_info["time"] = timesofar(self.ti)
            t1 = round(time.time() - self.ti, 0)
            job_info["time_in_s"] = t1
            diff_info["diff"][diff_key]["created_at"] = datetime.now()
        if "diff" in extra:
            diff_info["diff"][diff_key].update(extra["diff"])
        if "job" in extra:
            job_info.update(extra["job"])
        # since the base is the merged collection, we register info there
        # as the new collection (diff results are associated to the most recent colleciton)
        build = src_build.find_one({'_id': self.new.target_name})
        if not build:
            self.logger.info(
                "Can't find build document '%s', no status to register" %
                self.new.target_name)
            return
        if init:
            # init timer for this step
            self.ti = time.time()
            src_build.update({'_id': self.new.target_name},
                             {"$push": {
                                 'jobs': job_info
                             }})
            # now refresh/sync
            build = src_build.find_one({'_id': self.new.target_name})
        else:
            # merge extra at root level
            # (to keep building data...) and update the last one
            # (it's been properly created before when init=True)
            build["jobs"] and build["jobs"][-1].update(job_info)

            def merge_info(target, d):
                if "__REPLACE__" in d.keys():
                    d.pop("__REPLACE__")
                    target = d
                else:
                    for k, v in d.items():
                        if type(v) == dict:
                            # only merge if both are dict (otherwise replace/set with d)
                            if k in target and type(target[k]) == dict:
                                target[k] = merge_info(target[k], v)
                            else:
                                v.pop("__REPLACE__", None)
                                # merge v with "nothing" just to make sure to remove any "__REPLACE__"
                                v = merge_info({}, v)
                                target[k] = v
                        else:
                            target[k] = v
                return target

            build = merge_info(build, diff_info)
            src_build.replace_one({"_id": build["_id"]}, build)

    @asyncio.coroutine
    def diff_cols(self,
                  old_db_col_names,
                  new_db_col_names,
                  batch_size,
                  steps,
                  mode=None,
                  exclude=[]):
        """
        Compare new with old collections and produce diff files. Root keys can be excluded from
        comparison with "exclude" parameter

        `*_db_col_names` can be:
          1. a colleciton name (as a string) asusming they are
             in the target database.
          2. tuple with 2 elements, the first one is then either "source" or "target"
             to respectively specify src or target database, and the second element is
             the collection name.
          3. tuple with 3 elements (URI,db,collection), looking like:
             ("mongodb://user:pass@host","dbname","collection"), allowing to specify
             any connection on any server

        steps: - 'content' will perform diff on actual content.
               - 'mapping' will perform diff on ES mappings (if target collection involved)
               - 'reduce' will merge diff files, trying to avoid having many small files
               - 'post' is a hook to do stuff once everything is merged (override method post_diff_cols)

        mode: 'purge' will remove any existing files for this comparison while 'resume' will happily ignore
              existing data and to whatever it's requested (like running steps="post" on existing diff folder...)
        """
        # these ones are used to point to the build doc, not the underlying backned
        # (ie. if link builder has been used, it refers a colleciton in src_db, but
        # we need the metadata from the build doc too)
        self.new = create_backend(new_db_col_names, follow_ref=False)
        self.old = create_backend(old_db_col_names, follow_ref=False)
        # these point to the actual collection containing data
        content_new = create_backend(new_db_col_names, follow_ref=True)
        content_old = create_backend(old_db_col_names, follow_ref=True)
        # check what to do
        if type(steps) == str:
            steps = [steps]

        diff_folder = generate_folder(btconfig.DIFF_PATH, old_db_col_names,
                                      new_db_col_names)

        if mode != "force" and os.path.exists(
                diff_folder) and "content" in steps:
            if mode == "purge" and os.path.exists(diff_folder):
                rmdashfr(diff_folder)
            elif mode != "resume":
                raise FileExistsError(
                    "Found existing files in '%s', use mode='purge'" %
                    diff_folder)
        if not os.path.exists(diff_folder):
            os.makedirs(diff_folder)

        # create metadata file storing info about how we created the diff
        # and some summary data
        diff_stats = {
            "update": 0,
            "add": 0,
            "delete": 0,
            "mapping_changed": False
        }
        self.metadata_filename = os.path.join(diff_folder, "metadata.json")
        if os.path.exists(self.metadata_filename):
            # load previous metadata in case we try to update/complete diff
            self.metadata = json.load(open(self.metadata_filename))
        else:
            assert self.old.version, "Version for 'old' collection not defined"
            self.metadata = {
                "diff": {
                    "type": self.diff_type,
                    "func": self.diff_func.__name__,
                    "version": "%s.%s" % (self.old.version, self.new.version),
                    "stats": diff_stats,  # ref to diff_stats
                    "files": [],
                    # when "new" is a target collection:
                    "mapping_file": None,
                    "info": {
                        "generated_on": str(datetime.now()),
                        "exclude": exclude,
                        "steps": steps,
                        "mode": mode,
                        "batch_size": batch_size
                    }
                },
                "old": {
                    "backend": old_db_col_names,
                    "version": self.old.version
                },
                "new": {
                    "backend": new_db_col_names,
                    "version": self.new.version
                },
                # when "new" is a mongodb target collection:
                "_meta": {},
                "build_config": {},
            }
        if isinstance(
                self.new, DocMongoBackend
        ) and self.new.target_collection.database.name == btconfig.DATA_TARGET_DATABASE:
            new_doc = get_src_build().find_one(
                {"_id": self.new.target_collection.name})
            if not new_doc:
                raise DifferException("Collection '%s' has no corresponding build document" %
                                      self.new.target_collection.name)
            self.metadata["_meta"] = self.get_metadata()
            self.metadata["build_config"] = new_doc.get("build_config")

        got_error = False
        if "mapping" in steps:

            def diff_mapping(old, new, diff_folder):
                summary = {}
                old_build = get_src_build().find_one(
                    {"_id": old.target_collection.name})
                new_build = get_src_build().find_one(
                    {"_id": new.target_collection.name})
                if old_build and new_build:
                    # mapping diff always in jsondiff
                    mapping_diff = jsondiff(old_build["mapping"],
                                            new_build["mapping"])
                    if mapping_diff:
                        file_name = os.path.join(diff_folder, "mapping.pyobj")
                        dump(mapping_diff, file_name)
                        md5 = md5sum(file_name)
                        summary["mapping_file"] = {
                            "name": os.path.basename(file_name),
                            "md5sum": md5,
                            "size": os.stat(file_name).st_size
                        }
                else:
                    self.logger.info("Neither '%s' nor '%s' have mappings associated to them, skip" %
                                     (old.target_collection.name, new.target_collection.name))
                return summary

            def mapping_diffed(f):
                res = f.result()
                self.register_status("success", job={"step": "diff-mapping"})
                if res.get("mapping_file"):
                    nonlocal got_error
                    # check mapping differences: only "add" ops are allowed, as any others actions would be
                    # ingored by ES once applied (you can't update/delete elements of an existing mapping)
                    mf = os.path.join(diff_folder, res["mapping_file"]["name"])
                    ops = loadobj(mf)
                    for op in ops:
                        if op["op"] != "add":
                            err = DifferException("Found diff operation '%s' in mapping file, " % op["op"]
                                                  + " only 'add' operations are allowed. You can still produce the "
                                                  + "diff by removing 'mapping' from 'steps' arguments. "
                                                  + "Ex: steps=['content']. Diff operation was: %s" % op)
                            got_error = err
                    self.metadata["diff"]["mapping_file"] = res["mapping_file"]
                    diff_stats["mapping_changed"] = True
                    json.dump(self.metadata,
                              open(self.metadata_filename, "w"),
                              indent=True)

                self.logger.info(
                    "Diff file containing mapping differences generated: %s" %
                    res.get("mapping_file"))

            pinfo = self.get_pinfo()
            pinfo["source"] = "%s vs %s" % (self.new.target_name,
                                            self.old.target_name)
            pinfo["step"] = "mapping: old vs new"
            self.register_status("diffing",
                                 transient=True,
                                 init=True,
                                 job={"step": "diff-mapping"})
            job = yield from self.job_manager.defer_to_thread(
                pinfo, partial(diff_mapping, self.old, self.new, diff_folder))
            job.add_done_callback(mapping_diffed)
            yield from job
            if got_error:
                raise got_error

        if content_old == content_new:
            self.logger.info(
                "Old and new collections are the same, skipping 'content' step"
            )
        elif "content" in steps:
            cnt = 0
            jobs = []
            pinfo = self.get_pinfo()
            pinfo["source"] = "%s vs %s" % (content_new.target_name,
                                            content_old.target_name)
            pinfo["step"] = "content: new vs old"
            data_new = id_feeder(content_new, batch_size=batch_size)
            selfcontained = "selfcontained" in self.diff_type
            self.register_status("diffing",
                                 transient=True,
                                 init=True,
                                 job={"step": "diff-content"})
            for id_list_new in data_new:
                cnt += 1
                pinfo["description"] = "batch #%s" % cnt

                def diffed(f):
                    res = f.result()
                    diff_stats["update"] += res["update"]
                    diff_stats["add"] += res["add"]
                    if res.get("diff_file"):
                        self.metadata["diff"]["files"].append(res["diff_file"])
                    self.logger.info("(Updated: {}, Added: {})".format(
                        res["update"], res["add"]))
                    self.register_status("success",
                                         job={"step": "diff-content"})

                self.logger.info("Creating diff worker for batch #%s" % cnt)
                job = yield from self.job_manager.defer_to_process(
                    pinfo,
                    partial(diff_worker_new_vs_old, id_list_new,
                            old_db_col_names, new_db_col_names, cnt,
                            diff_folder, self.diff_func, exclude,
                            selfcontained))
                job.add_done_callback(diffed)
                jobs.append(job)
            yield from asyncio.gather(*jobs)
            self.logger.info(
                "Finished calculating diff for the new collection. Total number of docs updated: {}, added: {}"
                .format(diff_stats["update"], diff_stats["add"]))

            data_old = id_feeder(content_old, batch_size=batch_size)
            jobs = []
            pinfo = self.get_pinfo()
            pinfo["source"] = "%s vs %s" % (content_old.target_name,
                                            content_new.target_name)
            pinfo["step"] = "content: old vs new"
            for id_list_old in data_old:
                cnt += 1
                pinfo["description"] = "batch #%s" % cnt

                def diffed(f):
                    res = f.result()
                    diff_stats["delete"] += res["delete"]
                    if res.get("diff_file"):
                        self.metadata["diff"]["files"].append(res["diff_file"])
                    self.logger.info("(Deleted: {})".format(res["delete"]))

                self.logger.info("Creating diff worker for batch #%s" % cnt)
                job = yield from self.job_manager.defer_to_process(
                    pinfo,
                    partial(diff_worker_old_vs_new, id_list_old,
                            new_db_col_names, cnt, diff_folder))
                job.add_done_callback(diffed)
                jobs.append(job)
            yield from asyncio.gather(*jobs)
            self.logger.info(
                "Finished calculating diff for the old collection. Total number of docs deleted: {}"
                .format(diff_stats["delete"]))
            json.dump(self.metadata,
                      open(self.metadata_filename, "w"),
                      indent=True)

        self.logger.info(
            "Summary: (Updated: {}, Added: {}, Deleted: {}, Mapping changed: {})"
            .format(diff_stats["update"], diff_stats["add"],
                    diff_stats["delete"], diff_stats["mapping_changed"]))

        if "reduce" in steps:

            @asyncio.coroutine
            def merge_diff():
                self.logger.info("Reduce/merge diff files")
                max_diff_size = getattr(btconfig, "MAX_DIFF_SIZE",
                                        10 * 1024**2)
                current_size = 0
                cnt = 0
                final_res = []
                tomerge = []
                # .done contains original diff files
                done_folder = os.path.join(diff_folder, ".done")
                try:
                    os.mkdir(done_folder)
                except FileExistsError:
                    pass

                def merged(f, cnt):
                    nonlocal got_error
                    nonlocal final_res
                    try:
                        res = f.result()
                        final_res.extend(res)
                        self.logger.info("Diff file #%s created" % cnt)
                    except Exception as e:
                        got_error = e

                diff_files = [f for f in glob.glob(os.path.join(diff_folder, "*.pyobj"))
                              if not os.path.basename(f).startswith("mapping")]
                self.logger.info("%d diff files to process in total" %
                                 len(diff_files))
                jobs = []
                while diff_files:
                    if len(diff_files) % 100 == 0:
                        self.logger.info("%d diff files to process" %
                                         len(diff_files))
                    if current_size > max_diff_size:
                        job = yield from self.job_manager.defer_to_process(
                            pinfo,
                            partial(reduce_diffs, tomerge, cnt, diff_folder,
                                    done_folder))
                        job.add_done_callback(partial(merged, cnt=cnt))
                        jobs.append(job)
                        current_size = 0
                        cnt += 1
                        tomerge = []
                    else:
                        diff_file = diff_files.pop()
                        current_size += os.stat(diff_file).st_size
                        tomerge.append(diff_file)

                assert not diff_files

                if tomerge:
                    job = yield from self.job_manager.defer_to_process(
                        pinfo,
                        partial(reduce_diffs, tomerge, cnt, diff_folder,
                                done_folder))
                    job.add_done_callback(partial(merged, cnt=cnt))
                    jobs.append(job)
                    yield from job

                yield from asyncio.gather(*jobs)

                return final_res

            pinfo = self.get_pinfo()
            pinfo["source"] = "diff_folder"
            pinfo["step"] = "reduce"
            #job = yield from self.job_manager.defer_to_thread(pinfo,merge_diff)
            self.register_status("diffing",
                                 transient=True,
                                 init=True,
                                 job={"step": "diff-reduce"})
            res = yield from merge_diff()
            self.metadata["diff"]["files"] = res
            json.dump(self.metadata,
                      open(self.metadata_filename, "w"),
                      indent=True)
            if got_error:
                self.logger.exception("Failed to reduce diff files: %s" %
                                      got_error,
                                      extra={"notify": True})
                raise got_error
            self.register_status("success", job={"step": "diff-reduce"})

        if "post" in steps:
            pinfo = self.get_pinfo()
            pinfo["source"] = "diff_folder"
            pinfo["step"] = "post"
            self.register_status("diffing",
                                 transient=True,
                                 init=True,
                                 job={"step": "diff-post"})
            job = yield from self.job_manager.defer_to_thread(
                pinfo,
                partial(self.post_diff_cols,
                        old_db_col_names,
                        new_db_col_names,
                        batch_size,
                        steps,
                        mode=mode,
                        exclude=exclude))

            def posted(f):
                nonlocal got_error
                try:
                    res = f.result()
                    self.register_status("success",
                                         job={"step": "diff-post"},
                                         diff={"post": res})
                    self.logger.info("Post diff process successfully run: %s" %
                                     res)
                except Exception as e:
                    got_error = e

            job.add_done_callback(posted)
            yield from job
            json.dump(self.metadata,
                      open(self.metadata_filename, "w"),
                      indent=True)
            if got_error:
                self.logger.exception("Failed to run post diff process: %s" %
                                      got_error,
                                      extra={"notify": True})
                raise got_error

        strargs = "[old=%s,new=%s,steps=%s,diff_stats=%s]" % (
            old_db_col_names, new_db_col_names, steps, diff_stats)
        self.logger.info("success %s" % strargs, extra={"notify": True})

        # remove some metadata key for diff registering (some are already in build doc, it would be duplication)
        self.metadata.pop("_meta", None)
        self.metadata.pop("build_config", None)
        # record diff_folder so it's available for later without re-computing it
        self.metadata["diff_folder"] = diff_folder
        self.register_status("success", diff=self.metadata)
        return diff_stats

    def diff(self,
             old_db_col_names,
             new_db_col_names,
             batch_size=100000,
             steps=["content", "mapping", "reduce", "post"],
             mode=None,
             exclude=[]):
        """wrapper over diff_cols() coroutine, return a task"""
        job = asyncio.ensure_future(
            self.diff_cols(old_db_col_names, new_db_col_names, batch_size,
                           steps, mode, exclude))
        return job

    def get_metadata(self):
        new_doc = get_src_build().find_one(
            {"_id": self.new.target_collection.name})
        if not new_doc:
            raise DifferException("Collection '%s' has no corresponding build document" %
                                  self.new.target_collection.name)
        return new_doc.get("_meta", {})

    def post_diff_cols(self,
                       old_db_col_names,
                       new_db_col_names,
                       batch_size,
                       steps,
                       mode=None,
                       exclude=[]):
        """Post diff process hook. This coroutine will in a dedicated thread"""
        return


class ColdHotDiffer(BaseDiffer):
    @asyncio.coroutine
    def diff_cols(self, old_db_col_names, new_db_col_names, *args, **kwargs):
        self.new = create_backend(new_db_col_names)
        new_doc = get_src_build().find_one(
            {"_id": self.new.target_collection.name})
        assert "cold_collection" in new_doc.get("build_config", {}), "%s document doesn't have " % self.new.target_collection.name \
                                                                     + "a premerge collection declared. Is it really a hot merges collection ?"
        self.cold = create_backend(new_doc["build_config"]["cold_collection"])
        return super(ColdHotDiffer,
                     self).diff_cols(old_db_col_names, new_db_col_names, *args,
                                     **kwargs)

    def get_metadata(self):
        new_doc = get_src_build().find_one(
            {"_id": self.new.target_collection.name})
        cold_doc = get_src_build().find_one(
            {"_id": self.cold.target_collection.name})
        if not new_doc:
            raise DifferException("Collection '%s' has no corresponding build document" %
                                  self.new.target_collection.name)
        if not cold_doc:
            raise DifferException("Collection '%s' has no corresponding build document" %
                                  self.cold.target_collection.name)
        return merge_src_build_metadata([cold_doc, new_doc])


class ColdHotJsonDifferBase(ColdHotDiffer):
    def post_diff_cols(self,
                       old_db_col_names,
                       new_db_col_names,
                       batch_size,
                       steps,
                       mode=None,
                       exclude=[]):
        """
        Post-process the diff files by adjusting some jsondiff operation. Here's the process.
        For updated documents, some operations might illegal in the context of cold/hot merged collections.
        Case #1: "remove" op in an update
            from a cold/premerge collection, we have that doc:
                coldd = {"_id":1, "A":"123", "B":"456", "C":True}

            from the previous hot merge we have this doc:
                prevd = {"_id":1, "D":"789", "C":True, "E":"abc"}

            At that point, the final document, fully merged and indexed is:
                finald = {"_id":1, "A":"123", "B":"456", "C":True, "D":"789", "E":"abc"}
            We can notice field "C" is common to coldd and prevd.

            from the new hot merge, we have:
                newd = {"_id":1, "E","abc"} # C and D don't exist anymore

            Diffing prevd vs. newd will give jssondiff operations:
                [{'op': 'remove', 'path': '/C'}, {'op': 'remove', 'path': '/D'}]

            The problem here is 'C' is removed while it was already in cold merge, it should stay because it has come
            with some resource involved in the premerge (dependent keys, eg. myvariant, "observed" key comes with certain sources)
            => the jsondiff opetation on "C" must be discarded.

            Note: If operation involved a root key (not '/a/c' for instance) and if that key is found in the premerge, then
               then remove the operation. (note we just consider root keys, if the deletion occurs deeper in the document,
               it's just a legal operation updating innder content)

        For deleted documents, the same kind of logic applies
        Case #2: "delete"
            from a cold/premerge collection, we have that doc:
                coldd = {"_id":1, "A":"123", "B":"456", "C":True}
            from the previous hot merge we have this doc:
                prevd = {"_id":1, "D":"789", "C":True}
            fully merged doc:
                finald = {"_id":1, "A":"123", "B":"456", "C":True, "D":"789"}
            from the new hot merge, we have:
                newd = {} # document doesn't exist anymore

            Diffing prevd vs. newd will mark document with _id == 1 to be deleted
            The problem is we have data for _id=1 on the premerge collection, if we delete the whole document we'd loose too much
            information.
            => the deletion must converted into specific "remove" jsondiff operations, for the root keys found in prevd on not in coldd
               (in that case: [{'op':'remove', 'path':'/D'}], and not "C" as C is in premerge)
        """
        # we should be able to find a cold_collection definition in the src_build doc
        # and it should be the same for both old and new
        old_doc = get_src_build().find_one({"_id": old_db_col_names})
        new_doc = get_src_build().find_one({"_id": new_db_col_names})
        assert "build_config" in old_doc and "cold_collection" in old_doc["build_config"], \
            "No cold collection defined in src_build for %s" % old_db_col_names
        assert "build_config" in new_doc and "cold_collection" in new_doc["build_config"], \
            "No cold collection defined in src_build for %s" % new_db_col_names
        assert old_doc["build_config"]["cold_collection"] == new_doc["build_config"]["cold_collection"], \
            "Cold collections are different in src_build docs %s and %s" % (old_db_col_names, new_db_col_names)
        coldcol = get_target_db()[new_doc["build_config"]["cold_collection"]]
        assert coldcol.count() > 0, "Cold collection is empty..."
        diff_folder = generate_folder(btconfig.DIFF_PATH, old_db_col_names,
                                      new_db_col_names)
        diff_files = glob.glob(os.path.join(diff_folder, "diff_*.pyobj"))
        fixed = 0
        for diff_file in diff_files:
            dirty = False
            self.logger.info("Post-processing diff file %s" % diff_file)
            data = loadobj(diff_file)
            # update/remove case #1
            for updt in data["update"]:
                toremove = []
                for patch in updt["patch"]:
                    pathk = patch["path"].split("/")[
                        1:]  # remove / at the beginning of the path
                    if patch["op"] == "remove" and \
                            len(pathk) == 1:
                        # let's query the premerge
                        coldd = coldcol.find_one({"_id": updt["_id"]})
                        if coldd and pathk[0] in coldd:
                            self.logger.info(
                                "Fixed a root key in cold collection that should be preserved: '%s' (for doc _id '%s')"
                                % (pathk[0], updt["_id"]))
                            toremove.append(patch)
                            fixed += 1
                            dirty = True
                for p in toremove:
                    updt["patch"].remove(p)
            # delete case #2
            toremove = []
            prevcol = get_target_db()[old_doc["target_name"]]
            for delid in data["delete"]:
                coldd = coldcol.find_one({"_id": delid})
                if not coldd:
                    # true deletion is required
                    continue
                else:
                    prevd = prevcol.find_one({"_id": delid})
                    prevs = set(prevd.keys())
                    colds = set(coldd.keys())
                    keys = prevs.difference(
                        colds
                    )  # keys exclusively in prevd that should be removed
                    patches = []
                    for k in keys:
                        patches.append({"op": "remove", "path": "/%s" % k})
                    data["update"].append({"_id": delid, "patch": patches})
                    self.logger.info(
                        "Fixed a delete document by converting to update/remove jsondiff operations for keys: %s (_id: '%s')"
                        % (keys, delid))
                    fixed += 1
                    dirty = True
                    toremove.append(delid)
            for i in toremove:
                data["delete"].remove(i)

            if dirty:
                dump(data, diff_file, compress="lzma")
                name = os.path.basename(diff_file)
                md5 = md5sum(diff_file)
                # find info to adjust md5sum
                found = False
                for i, df in enumerate(self.metadata["diff"]["files"]):
                    if df["name"] == name:
                        found = True
                        break
                assert found, "Couldn't find file information in metadata (with md5 value), try to rebuild_diff_file_list() ?"
                size = os.stat(diff_file).st_size
                self.metadata["diff"]["files"][i] = {
                    "name": name,
                    "md5sum": md5,
                    "size": size
                }
                self.logger.info(self.metadata["diff"]["files"])

        self.logger.info(
            "Post-diff process fixing jsondiff operations done: %s fixed" %
            fixed)
        return {"fixed": fixed}


class JsonDiffer(BaseDiffer):
    diff_type = "jsondiff"

    def __init__(self, diff_func=diff_docs_jsonpatch, *args, **kwargs):
        super(JsonDiffer, self).__init__(diff_func=diff_func, *args, **kwargs)


class SelfContainedJsonDiffer(JsonDiffer):
    diff_type = "jsondiff-selfcontained"


class ColdHotJsonDiffer(ColdHotJsonDifferBase, JsonDiffer):
    diff_type = "coldhot-jsondiff"


class ColdHotSelfContainedJsonDiffer(ColdHotJsonDifferBase,
                                     SelfContainedJsonDiffer):
    diff_type = "coldhot-jsondiff-selfcontained"


def diff_worker_new_vs_old(id_list_new,
                           old_db_col_names,
                           new_db_col_names,
                           batch_num,
                           diff_folder,
                           diff_func,
                           exclude=[],
                           selfcontained=False):
    new = create_backend(new_db_col_names, follow_ref=True)
    old = create_backend(old_db_col_names, follow_ref=True)
    docs_common = old.mget_from_ids(id_list_new)
    ids_common = [_doc['_id'] for _doc in docs_common]
    id_in_new = list(set(id_list_new) - set(ids_common))
    _updates = []
    if len(ids_common) > 0:
        _updates = diff_func(old, new, list(ids_common), exclude_attrs=exclude)
    file_name = os.path.join(diff_folder, "%s.pyobj" % str(batch_num))
    _result = {
        'add': id_in_new,
        'update': _updates,
        'delete': [],
        'source': new.target_name,
        'timestamp': get_timestamp()
    }
    if selfcontained:
        # consume generator as result will be pickled
        _result["add"] = [d for d in new.mget_from_ids(id_in_new)]
    summary = {"add": len(id_in_new), "update": len(_updates), "delete": 0}
    if len(_updates) != 0 or len(id_in_new) != 0:
        dump(_result, file_name)
        # compute md5 so when downloaded, users can check integreity
        md5 = md5sum(file_name)
        summary["diff_file"] = {
            "name": os.path.basename(file_name),
            "md5sum": md5,
            "size": os.stat(file_name).st_size
        }

    return summary


def diff_worker_old_vs_new(id_list_old, new_db_col_names, batch_num,
                           diff_folder):
    new = create_backend(new_db_col_names, follow_ref=True)
    docs_common = new.mget_from_ids(id_list_old)
    ids_common = [_doc['_id'] for _doc in docs_common]
    id_in_old = list(set(id_list_old) - set(ids_common))
    file_name = os.path.join(diff_folder, "%s.pyobj" % str(batch_num))
    _result = {
        'delete': id_in_old,
        'add': [],
        'update': [],
        'source': new.target_name,
        'timestamp': get_timestamp()
    }
    summary = {"add": 0, "update": 0, "delete": len(id_in_old)}
    if len(id_in_old) != 0:
        dump(_result, file_name)
        # compute md5 so when downloaded, users can check integreity
        md5 = md5sum(file_name)
        summary["diff_file"] = {
            "name": os.path.basename(file_name),
            "md5sum": md5,
            "size": os.stat(file_name).st_size
        }

    return summary


def diff_worker_count(id_list, db_col_names, batch_num):
    col = create_backend(db_col_names, follow_ref=True)
    docs = col.mget_from_ids(id_list)
    res = {}
    for doc in docs:
        for k in doc:
            res.setdefault(k, 0)
            res[k] += 1
    return res


class DiffReportRendererBase(object):
    def __init__(self,
                 max_reported_ids=None,
                 max_randomly_picked=None,
                 detailed=False):
        self.max_reported_ids = max_reported_ids or hasattr(btconfig, "MAX_REPORTED_IDS") and \
            btconfig.MAX_REPORTED_IDS or 1000
        self.max_randomly_picked = max_randomly_picked or hasattr(btconfig, "MAX_RANDOMLY_PICKED") and \
            btconfig.MAX_RANDOMLY_PICKED or 10
        self.detailed = detailed

    def save(self, report, filename):
        """
        Save report output (rendered) into filename
        """
        raise NotImplementedError("implement me")


class DiffReportTxt(DiffReportRendererBase):
    def save(self, report, filename="report.txt"):
        try:
            import prettytable
        except ImportError:
            raise ImportError(
                "Please install prettytable to use this rendered")

        def build_id_table(subreport):
            if self.detailed:
                table = prettytable.PrettyTable(["IDs", "Root keys"])
                table.align["IDs"] = "l"
                table.align["Root keys"] = "l"
            else:
                table = prettytable.PrettyTable(["IDs"])
                table.align["IDs"] = "l"
            if subreport["count"] <= self.max_reported_ids:
                ids = subreport["ids"]
            else:
                ids = [
                    random.choice(subreport["ids"])
                    for i in range(self.max_reported_ids)
                ]
            for dat in ids:
                if self.detailed:
                    # list of [_id,[keys]]
                    table.add_row([dat[0], ", ".join(dat[1])])
                else:
                    table.add_row([dat])

            return table

        txt = ""
        title = "Diff report (generated on %s)" % datetime.now()
        txt += title + "\n"
        txt += "".join(["="] * len(title)) + "\n"
        txt += "\n"
        txt += "Metadata\n"
        txt += "--------\n"
        if report.get("metadata", {}):
            txt += "Old collection: %s\n" % repr(report["metadata"].get("old"))
            txt += "New collection: %s\n" % repr(report["metadata"].get("new"))
            txt += "Batch size: %s\n" % report["metadata"]["diff"]["info"].get(
                "batch_size")
            txt += "Steps: %s\n" % report["metadata"]["diff"]["info"].get(
                "steps")
            txt += "Key(s) excluded: %s\n" % report["metadata"]["diff"][
                "info"].get("exclude")
            txt += "Diff generated on: %s\n" % report["metadata"]["diff"][
                "info"].get("generated_on")
        else:
            txt += "No metadata found in report\n"
        txt += "\n"
        txt += "Summary\n"
        txt += "-------\n"
        txt += "Added documents: %s\n" % report["added"]["count"]
        txt += "Deleted documents: %s\n" % report["deleted"]["count"]
        txt += "Updated documents: %s\n" % report["updated"]["count"]
        txt += "\n"
        root_keys = report.get("metadata",
                               {}).get("diff",
                                       {}).get("stats",
                                               {}).get("root_keys", {})
        if root_keys:
            for src in sorted(root_keys):
                txt += "%s: %s\n" % (src, root_keys[src])
        else:
            txt += "No root keys count found in report\n"
        txt += "\n"
        txt += "Added documents (%s randomly picked from report)\n" % self.max_reported_ids
        txt += "------------------------------------------------\n"
        if report["added"]["count"]:
            table = build_id_table(report["added"])
            txt += table.get_string()
            txt += "\n"
        else:
            txt += "No added document found in report\n"
        txt += "\n"
        txt += "Deleted documents (%s randomly picked from report)\n" % self.max_reported_ids
        txt += "--------------------------------------------------\n"
        if report["deleted"]["count"]:
            table = build_id_table(report["deleted"])
            txt += table.get_string()
            txt += "\n"
        else:
            txt += "No deleted document found in report\n"
        txt += "\n"
        txt += "Updated documents (%s examples randomly picked from report)\n" % self.max_randomly_picked
        txt += "-----------------------------------------------------------\n"
        txt += "\n"
        for op in sorted(report["updated"]):
            if op == "count":
                continue  # already displayed
            if report["updated"][op]:
                table = prettytable.PrettyTable([op, "Count", "Examples"])
                table.sortby = "Count"
                table.reversesort = True
                table.align[op] = "l"
                table.align["Count"] = "r"
                table.align["Examples"] = "l"
                for path in report["updated"][op]:
                    info = report["updated"][op][path]
                    row = [path, info["count"]]
                    if info["count"] <= self.max_randomly_picked:
                        row.append(", ".join(info["ids"]))
                    else:
                        row.append(", ".join([
                            random.choice(info["ids"])
                            for i in range(self.max_randomly_picked)
                        ]))
                    table.add_row(row)
                txt += table.get_string()
                txt += "\n"
            else:
                txt += "No content found for diff operation '%s'\n" % op
            txt += "\n"
        txt += "\n"

        with open(os.path.join(report["diff_folder"], filename), "w") as fout:
            fout.write(txt)

        return txt


class DifferManager(BaseManager):
    def __init__(self, poll_schedule=None, *args, **kwargs):
        """
        DifferManager deals with the different differ objects used to create and
        analyze diff between datasources.
        """
        super(DifferManager, self).__init__(*args, **kwargs)
        self.log_folder = btconfig.LOG_FOLDER
        self.timestamp = datetime.now()
        self.poll_schedule = poll_schedule
        self.setup_log()

    def clean_stale_status(self):
        src_build = get_src_build()
        for build in src_build.find():
            dirty = False
            for job in build.get("jobs", []):
                if job.get("status") == "diffing":
                    logging.warning(
                        "Found stale build '%s', marking diff status as 'canceled'"
                        % build["_id"])
                    job["status"] = "canceled"
                    dirty = True
            if dirty:
                src_build.replace_one({"_id": build["_id"]}, build)

    def register_differ(self, klass):
        if klass.diff_type is None:
            raise DifferException("diff_type must be defined in %s" % klass)
        self.register[klass.diff_type] = partial(
            klass,
            log_folder=btconfig.LOG_FOLDER,
            job_manager=self.job_manager)

    def configure(self, partial_differs=[JsonDiffer, SelfContainedJsonDiffer]):
        for pdiffer in partial_differs:
            self.register_differ(pdiffer)

    def setup_log(self):
        self.logger, self.logfile = get_logger('diffmanager')

    def get_predicates(self):
        def no_other_diffmanager_step_running(job_manager):
            """DiffManager deals with diff report, release note, publishing,
            none of them should run more than one at a time"""
            # Note: report output is part a publish_diff, release_note is impacted by diff content,
            # overall we keep things simple and don't allow more than one diff manager job to run
            # at the same time
            return len([
                j for j in job_manager.jobs.values()
                if j["category"] == DIFFMANAGER_CATEGORY
            ]) == 0

        return [no_other_diffmanager_step_running]

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": DIFFMANAGER_CATEGORY,
            "source": "",
            "step": "",
            "description": ""
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    def __getitem__(self, diff_type):
        """
        Return an instance of a builder for the build named 'build_name'
        Note: each call returns a different instance (factory call behind the scene...)
        """
        # we'll get a partial class but will return an instance
        pclass = BaseManager.__getitem__(self, diff_type)
        return pclass()

    def diff(self,
             diff_type,
             old,
             new,
             batch_size=100000,
             steps=["content", "mapping", "reduce", "post"],
             mode=None,
             exclude=["_timestamp"]):
        """
        Run a diff to compare old vs. new collections. using differ algorithm diff_type. Results are stored in
        a diff folder.
        Steps can be passed to choose what to do:
        - count: will count root keys in new collections and stores them as statistics.
        - content: will diff the content between old and new. Results (diff files) format depends on diff_type
        """
        # Note: _timestamp is excluded by default since it's an internal field (and exists in mongo doc,
        #       but not in ES "_source" document (there's a root timestamp under control of
        #       _timestamp : {enable:true} in mapping
        try:
            differ = self[diff_type]
            old = old or get_previous_collection(new)
            job = differ.diff(old,
                              new,
                              batch_size=batch_size,
                              steps=steps,
                              mode=mode,
                              exclude=exclude)

            def diffed(f):
                try:
                    _ = f.result()
                    set_pending_to_release_note(new)
                except Exception as e:
                    self.logger.error("Error during diff: %s" % e)
                    raise

            job.add_done_callback(diffed)
            return job
        except KeyError as e:
            raise DifferException("No such differ '%s' (error: %s)" %
                                  (diff_type, e))

    def diff_report(self,
                    old_db_col_names,
                    new_db_col_names,
                    report_filename="report.txt",
                    format="txt",
                    detailed=True,
                    max_reported_ids=None,
                    max_randomly_picked=None,
                    mode=None):

        max_reported_ids = max_reported_ids or hasattr(btconfig, "MAX_REPORTED_IDS") and \
            btconfig.MAX_REPORTED_IDS or 1000
        max_randomly_picked = max_randomly_picked or hasattr(btconfig, "MAX_RANDOMLY_PICKED") and \
            btconfig.MAX_RANDOMLY_PICKED or 10

        def do():
            if mode == "purge" or not os.path.exists(reportfilepath):
                assert format == "txt", "Only 'txt' format supported for now"
                report = self.build_diff_report(diff_folder, detailed,
                                                max_reported_ids)
                render = DiffReportTxt(max_reported_ids=max_reported_ids,
                                       max_randomly_picked=max_randomly_picked,
                                       detailed=detailed)
                return render.save(report, report_filename)
            else:
                self.logger.debug("Report already generated, now using it")
                return open(reportfilepath).read()

        @asyncio.coroutine
        def main(diff_folder):
            got_error = False
            pinfo = self.get_pinfo()
            pinfo["step"] = "report"
            pinfo["source"] = diff_folder
            pinfo["description"] = report_filename
            job = yield from self.job_manager.defer_to_thread(pinfo, do)

            def reported(f):
                nonlocal got_error
                try:
                    _ = f.result()
                    self.logger.info("Diff report ready, saved in %s" %
                                     reportfilepath,
                                     extra={
                                         "notify": True,
                                         "attach": reportfilepath
                                     })
                except Exception as e:
                    got_error = e

            job.add_done_callback(reported)
            yield from job
            if got_error:
                self.logger.exception("Failed to create diff report: %s" %
                                      got_error,
                                      extra={"notify": True})
                raise got_error

        diff_folder = generate_folder(btconfig.DIFF_PATH, old_db_col_names,
                                      new_db_col_names)
        reportfilepath = os.path.join(diff_folder, report_filename)
        job = asyncio.ensure_future(main(diff_folder))
        return job

    def build_diff_report(self,
                          diff_folder,
                          detailed=True,
                          max_reported_ids=None):
        """
        Analyze diff files in diff_folder and give a summy of changes.
        max_reported_ids is the number of IDs contained in the report for each part.
        detailed will trigger a deeper analysis, takes more time.
        """

        max_reported_ids = max_reported_ids or hasattr(btconfig, "MAX_REPORTED_IDS") and \
            btconfig.MAX_REPORTED_IDS or 1000

        update_details = {
            "add": {},  # "count": 0, "data": {} },
            "remove": {},  # "count": 0, "data": {} },
            "replace": {},  # "count": 0, "data": {} },
            "move": {},  # "count": 0, "data": {} },
            "count": 0,
        }
        adds = {"count": 0, "ids": []}
        dels = {"count": 0, "ids": []}
        sources = {}

        if os.path.isabs(diff_folder):
            data_folder = diff_folder
        else:
            data_folder = os.path.join(btconfig.DIFF_PATH, diff_folder)

        metadata = {}
        try:
            metafile = os.path.join(data_folder, "metadata.json")
            metadata = json.load(open(metafile))
        except FileNotFoundError:
            logging.warning("Not metadata found in diff folder")
            if detailed:
                raise Exception(
                    "Can't perform detailed analysis without a metadata file")

        def analyze(diff_file, detailed):
            data = loadobj(diff_file)
            sources[data["source"]] = 1
            if detailed:
                # TODO: if self-contained, no db connection needed
                new_col = create_backend(metadata["new"]["backend"],
                                         follow_ref=True)
                old_col = create_backend(metadata["old"]["backend"],
                                         follow_ref=True)
            if len(adds["ids"]) < max_reported_ids:
                if detailed:
                    # look for which root keys were added in new collection
                    for _id in data["add"]:
                        # selfcontained = dict for whole doc (see TODO above)
                        if type(_id) == dict:
                            _id = _id["_id"]
                        doc = new_col.get_from_id(_id)
                        rkeys = sorted(doc.keys())
                        adds["ids"].append([_id, rkeys])
                else:
                    if data["add"] and type(data["add"][0]) == dict:
                        adds["ids"].extend([d["_id"] for d in data["add"]])
                    else:
                        adds["ids"].extend(data["add"])
            adds["count"] += len(data["add"])
            if len(dels["ids"]) < max_reported_ids:
                if detailed:
                    # look for which root keys were deleted in old collection
                    for _id in data["delete"]:
                        doc = old_col.get_from_id(_id)
                        rkeys = sorted(doc.keys())
                        dels["ids"].append([_id, rkeys])
                else:
                    dels["ids"].extend(data["delete"])
            dels["count"] += len(data["delete"])
            for up in data["update"]:
                for patch in up["patch"]:
                    update_details[patch["op"]].setdefault(
                        patch["path"], {
                            "count": 0,
                            "ids": []
                        })
                    if len(update_details[patch["op"]][patch["path"]]
                           ["ids"]) < max_reported_ids:
                        update_details[patch["op"]][
                            patch["path"]]["ids"].append(up["_id"])
                    update_details[patch["op"]][patch["path"]]["count"] += 1
            update_details["count"] += len(data["update"])

            assert len(
                sources
            ) == 1, "Should have one datasource from diff files, got: %s" % [
                s for s in sources
            ]

        # we randomize files order b/c we randomly pick some examples from those
        # files. If files contains data in order (like chrom 1, then chrom 2)
        # we won't have a representative sample
        files = glob.glob(os.path.join(data_folder, "*.pyobj"))
        random.shuffle(files)
        total = len(files)
        for i, f in enumerate(files):
            if os.path.basename(f).startswith("mapping"):
                logging.debug("Skip mapping file")
                continue
            logging.info("Running report worker for '%s' (%d/%d)" %
                         (f, i + 1, total))
            analyze(f, detailed)
        return {
            "added": adds,
            "deleted": dels,
            "updated": update_details,
            "diff_folder": diff_folder,
            "detailed": detailed,
            "metadata": metadata
        }

    def poll(self, state, func):
        super(DifferManager, self).poll(state, func, col=get_src_build())

    def trigger_diff(self, diff_type, doc, **kwargs):
        """
        Launch a diff given a src_build document. In order to
        know the first collection to diff against, get_previous_collection()
        method is used.
        """
        new_db_col_names = doc["_id"]
        old_db_col_names = get_previous_collection(new_db_col_names)
        self.diff(diff_type, old_db_col_names, new_db_col_names, **kwargs)

    def rebuild_diff_file_list(self, diff_folder):
        diff_files = glob.glob(os.path.join(diff_folder, "*.pyobj"))
        metadata = json.load(open(os.path.join(diff_folder, "metadata.json")))
        try:
            metadata["diff"]["files"] = []
            if "mapping_file" not in metadata["diff"]:
                metadata["diff"]["mapping_file"] = None
            for diff_file in diff_files:
                name = os.path.basename(diff_file)
                md5 = md5sum(diff_file)
                info = {
                    "md5sum": md5,
                    "name": name,
                    "size": os.stat(diff_file).st_size
                }
                if "mapping" in diff_file:  # dirty...
                    metadata["diff"]["mapping"] = info
                else:
                    metadata["diff"]["files"].append(info)
            json.dump(metadata,
                      open(os.path.join(diff_folder, "metadata.json"), "w"),
                      indent=True)
            self.logger.info(
                "Successfully rebuild diff_files list with all files found in %s"
                % diff_folder)
        except KeyError as e:
            self.logging.error(
                "Metadata is too much damaged, can't rebuild diff_files list: %s"
                % e)

    def diff_info(self):
        dtypes = self.register.keys()
        res = {}
        for typ in dtypes:
            res[typ] = {}
        return res


def reduce_diffs(diffs, num, diff_folder, done_folder):
    assert diffs
    res = []
    fn = "diff_%s.pyobj" % num
    logging.info("Merging %s => %s" % ([os.path.basename(f)
                                        for f in diffs], fn))

    if len(diffs) == 1:
        # just rename
        outf = os.path.join(diff_folder, fn)
        shutil.copyfile(diffs[0], outf)
        res.append({
            "name": fn,
            "md5sum": md5sum(outf),
            "size": os.stat(outf).st_size
        })
        os.rename(diffs[0],
                  os.path.join(done_folder, os.path.basename(diffs[0])))
        return res

    merged = loadobj(diffs[0])
    os.rename(diffs[0], os.path.join(done_folder, os.path.basename(diffs[0])))
    for diff_fn in diffs[1:]:
        diff = loadobj(diff_fn)
        assert merged["source"] == diff["source"]
        for k in ["add", "delete", "update"]:
            merged[k].extend(diff[k])
        os.rename(diff_fn, os.path.join(done_folder,
                                        os.path.basename(diff_fn)))
    dump(merged, os.path.join(diff_folder, fn), compress="lzma")
    file_name = os.path.join(diff_folder, fn)
    res.append({
        "name": fn,
        "md5sum": md5sum(file_name),
        "size": os.stat(file_name).st_size
    })
    return res


def set_pending_to_diff(col_name):
    src_build = get_src_build()
    src_build.update({"_id": col_name}, {"$addToSet": {"pending": "diff"}})



"""

Biothings Web API Handlers

    Supports: (all features above and)
    - payload type 'application/json' (through self.json_arguments)
    - parsing keyword argument options (type, default, alias, ...)
    - multi-type dictionary output (json, yaml, html, msgpack)
    - standardized error response (exception -> error template)
    - analytics and usage tracking (Google Analytics and AWS)
    - default common http headers (CORS and Cache Control)

    Subclasses:
    - discovery.web.api.APIBaseHandler
"""

import datetime
import json
from collections import OrderedDict
from pprint import pformat
from urllib.parse import (parse_qs, unquote_plus, urlencode, urlparse,
                          urlunparse)

from tornado.escape import json_decode
from tornado.web import HTTPError

from biothings.utils.common import DateTimeJSONEncoder
from biothings.utils.web.analytics import GAMixIn
from biothings.utils.web.tracking import StandaloneTrackingMixin
from biothings.web.options import OptionArgError

from . import BaseHandler
from .exceptions import BadRequest

try:
    import msgpack

    def msgpack_encode_datetime(obj):
        if isinstance(obj, datetime.datetime):
            return {'__datetime__': True,
                    'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f")}
        return obj
except ImportError:
    SUPPORT_MSGPACK = False
else:
    SUPPORT_MSGPACK = True


try:
    import yaml
except ImportError:
    SUPPORT_YAML = False
else:
    SUPPORT_YAML = True

__all__ = [
    'BaseAPIHandler',
    'APISpecificationHandler'
]

class BaseAPIHandler(BaseHandler, GAMixIn, StandaloneTrackingMixin):

    name = ''
    kwargs = None  # dict
    kwarg_groups = ()
    kwarg_methods = ()
    format = 'json'

    def initialize(self):

        self.args = {}  # processed args will be available here
        self.args_json = {}  # applicatoin/json type body
        self.event = {
            'category': '{}_api'.format(self.web_settings.API_VERSION),
            'action': self.request.method,  # 'query_get', 'fetch_all', etc.
            # 'label': 'total', 'qsize', etc.
            # 'value': 0, corresponds to label ...
        }

    def prepare(self):
        '''
        Extract body and url query parameters into functional groups.
        Typify predefined user inputs patterns here. Rules:

            * Inputs are combined and then separated into functional catagories.
            * Duplicated query or body arguments will overwrite the previous value.
            * JSON body input will not overwrite query arguments in URL.
            * Path arguments can overwirte all other existing values.

        Extend to add more customizations.
        '''
        if self.request.headers.get('Content-Type') == 'application/json':
            try:
                self.args_json = json_decode(self.request.body)
            except json.JSONDecodeError:
                raise HTTPError(400, reason='Invalid JSON body.')

        args = dict(self.args_json)
        args.update({key: self.get_argument(key)
                     for key in self.request.arguments})

        self.logger.debug("Incoming Requests:\n%s\n%s",
                          self.request.uri,
                          pformat(args, width=150))

        if self.name:
            options = self.web_settings.optionsets.get(self.name)
            try:
                self.args = options.parse(
                    self.request.method, args,
                    self.path_args, self.path_kwargs)
            except OptionArgError as err:
                raise BadRequest(**err.info)

        self.logger.debug("Processed Arguments:\n%s", pformat(self.args, width=150))

    def write(self, chunk):
        """
        Override to write output basing on the specified format.
        """
        if isinstance(chunk, dict) and self.format not in ('json', ''):
            if self.format == 'yaml' and SUPPORT_YAML:
                self.set_header("Content-Type", "text/x-yaml; charset=UTF-8")
                chunk = self._format_yaml(chunk)

            elif self.format == 'msgpack' and SUPPORT_MSGPACK:
                self.set_header("Content-Type", "application/x-msgpack")
                chunk = self._format_msgpack(chunk)

            elif self.format == 'html':
                self.set_header("Content-Type", "text/html; charset=utf-8")
                chunk = self._format_html(chunk)
            else:
                self.set_status(400)
                chunk = {"success": False, "code": 400,
                         "error": f"Server not configured to output {self.format}."}

        elif isinstance(chunk, (dict, list)):
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            chunk = json.dumps(chunk, cls=DateTimeJSONEncoder)

        super().write(chunk)

    def _format_yaml(self, data):

        def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
            class OrderedDumper(Dumper):
                pass

            def _dict_representer(dumper, data):
                return dumper.represent_mapping(
                    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                    data.items())
            OrderedDumper.add_representer(OrderedDict, _dict_representer)
            return yaml.dump(data, stream, OrderedDumper, **kwds)

        return ordered_dump(
            data=data, Dumper=yaml.SafeDumper, default_flow_style=False)

    def _format_msgpack(self, data):

        return msgpack.packb(
            data, use_bin_type=True, default=msgpack_encode_datetime)

    def _format_html(self, data):

        _link = self.request.full_url()
        d = urlparse(_link)
        q = parse_qs(d.query)
        q.pop('format', None)
        d = d._replace(query=urlencode(q, True))
        _link = urlunparse(d)

        return self.render_string(
            template_name="api.html",
            data=json.dumps(data),
            img_src=self.web_settings.HTML_OUT_HEADER_IMG,
            link=_link,  # url to get regular format
            link_decode=unquote_plus(_link),
            title_html=self.web_settings.HTML_OUT_TITLE,
            docs_link=getattr(self.web_settings, self.name.upper()+'_DOCS_URL', '')
        )

    def on_finish(self):
        """
        This is a tornado lifecycle hook.
        Override to provide tracking features.
        """
        self.logger.debug("Event: %s", self.event)
        self.ga_track(self.event)
        self.self_track(self.event)

    def write_error(self, status_code, **kwargs):

        reason = kwargs.pop('reason', self._reason)
        assert '\n' not in reason

        message = {
            "code": status_code,
            "success": False,
            "error": reason
        }
        if 'exc_info' in kwargs:
            exception = kwargs['exc_info'][1]
            message.update(self.parse_exception(exception))

        self.finish(message)

    def parse_exception(self, exception):
        """
        Return customized error message basing on exception types.
        """
        if isinstance(exception, BadRequest):
            if exception.kwargs:
                return exception.kwargs
        return {}

    def options(self, *args, **kwargs):

        self.set_status(204)
        self.finish()

    def set_default_headers(self):

        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods",
                        self.web_settings.ACCESS_CONTROL_ALLOW_METHODS)
        self.set_header("Access-Control-Allow-Headers",
                        self.web_settings.ACCESS_CONTROL_ALLOW_HEADERS)
        self.set_header("Access-Control-Allow-Credentials", "false")
        self.set_header("Access-Control-Max-Age", "60")

        if not self.web_settings.DISABLE_CACHING:
            seconds = self.web_settings.CACHE_MAX_AGE
            self.set_header("Cache-Control", f"max-age={seconds}, public")

class APISpecificationHandler(BaseAPIHandler):

    def get(self):
        self.finish(self.web_settings.optionsets.log())



import os
import sys
import types
import logging
from pprint import pformat
import importlib

from biothings.utils.hub_db import get_data_plugin
from biothings.utils.manager import BaseSourceManager
from biothings.utils.hub_db import get_src_master, get_src_dump


class SourceManager(BaseSourceManager):
    """
    Helper class to get information about a datasource,
    whether it has a dumper and/or uploaders associated.
    """
    def __init__(self, source_list, dump_manager, upload_manager,
                 data_plugin_manager):
        self._orig_source_list = source_list
        self.source_list = None
        self.dump_manager = dump_manager
        self.upload_manager = upload_manager
        self.data_plugin_manager = data_plugin_manager
        self.reload()
        self.src_master = get_src_master()
        self.src_dump = get_src_dump()
        # honoring BaseSourceManager interface (gloups...-
        self.register = {}

    def reload(self):
        # clear registers
        self.dump_manager.register.clear()
        self.upload_manager.register.clear()
        # re-eval source list (so if it's a string, it'll re-discover sources)
        self.source_list = self.find_sources(self._orig_source_list)
        self.dump_manager.register_sources(self.source_list)
        self.upload_manager.register_sources(self.source_list)

    def find_sources(self, paths):
        sources = []

        if not type(paths) == list:
            paths = [paths]

        def eval_one_source(one_path):
            if "/" in one_path:
                # it's path to directory
                # expecting
                if one_path not in sys.path:
                    logging.info("Adding '%s' to python path" % one_path)
                    sys.path.insert(0, one_path)
                for d in os.listdir(one_path):
                    if d.endswith("__pycache__"):
                        continue
                    sources.append(d)
            else:
                # assuming it's path to a python module (oath.to.module)
                sources.append(one_path)

        def eval_one_root(root):
            logging.debug("Discovering sources in %s" % root)
            # root is a module path where sources can be found
            rootdir, __init__ = os.path.split(root.__file__)
            for srcdir in os.listdir(rootdir):
                if srcdir.endswith("__pycache__"):
                    continue
                srcpath = os.path.join(rootdir, srcdir)
                if os.path.isdir(srcpath):
                    srcmod_str = "%s.%s" % (root.__name__, srcdir)
                    sources.append(srcmod_str)

        for path in paths:
            if type(path) == str:
                eval_one_source(path)
            elif isinstance(path, types.ModuleType):
                eval_one_root(path)

        # clean with only those which can be imported
        sources = set(sources)
        for s in [s for s in sources]:
            try:
                importlib.import_module(s)
            except Exception as e:
                logging.error("Failed to discover source '%s': %s" % (s, e))
                sources.remove(s)

        logging.info("Found sources: %s" % sorted(sources))
        return sources

    def set_mapping_src_meta(self, subsrc, mini):
        # get mapping from uploader klass first (hard-coded), then src_master (generated/manual)
        src_meta = {}
        mapping = {}
        origin = None
        try:
            upk = self.upload_manager["%s.%s" % (mini["_id"], subsrc)]
            assert len(
                upk) == 1, "More than 1 uploader found, can't handle that..."
            upk = upk.pop()
            src_meta = upk.__metadata__["src_meta"]
            mapping = upk.get_mapping()
            origin = "uploader"
            if not mapping:
                raise AttributeError("Not hard-coded mapping")
        except (IndexError, KeyError, AttributeError) as e:
            logging.debug(
                "Can't find hard-coded mapping, now searching src_master: %s" %
                e)
            m = self.src_master.find_one({"_id": subsrc})
            mapping = m and m.get("mapping")
            origin = "master"
            # use metadata from upload or reconstitute(-ish)
            src_meta = src_meta or m and dict([(k, v) for (k, v) in m.items() if k not in ["_id", "name", "timestamp", "mapping"]])
        if mapping:
            mini.setdefault("mapping",
                            {}).setdefault(subsrc,
                                           {}).setdefault("mapping", mapping)
            mini.setdefault("mapping",
                            {}).setdefault(subsrc,
                                           {}).setdefault("origin", origin)
        if src_meta:
            mini.setdefault("__metadata__", {}).setdefault(subsrc, src_meta)

    def sumup_source(self, src, detailed=False):
        """Return minimal info about src"""

        mini = {}
        mini["_id"] = src.get("_id", src["name"])
        mini["name"] = src["name"]

        if src.get("download"):
            mini["download"] = {
                "status": src["download"].get("status"),
                "time": src["download"].get("time"),
                "started_at": src["download"].get("started_at"),
                "release": src["download"].get("release"),
                "data_folder": src["download"].get("data_folder"),
            }
            mini["download"]["dumper"] = src["download"].get("dumper", {})
            if src["download"].get("err"):
                mini["download"]["error"] = src["download"]["err"]

        count = 0
        if src.get("upload"):
            mini["upload"] = {"sources": {}}
            for job, info in src["upload"]["jobs"].items():
                mini["upload"]["sources"][job] = {
                    "time": info.get("time"),
                    "status": info.get("status"),
                    "count": info.get("count"),
                    "started_at": info.get("started_at"),
                    "release": info.get("release"),
                    "data_folder": info.get("data_folder"),
                }
                if info.get("err"):
                    mini["upload"]["sources"][job]["error"] = info["err"]
                count += info.get("count") or 0
                if detailed:
                    self.set_mapping_src_meta(job, mini)
        if src.get("inspect"):
            mini["inspect"] = {"sources": {}}
            for job, info in src["inspect"]["jobs"].items():
                if not detailed:
                    # remove big inspect data but preserve inspect status/info and errors
                    mode_has_error = []
                    mode_ok = []
                    for mode in info.get("inspect", {}).get("results", {}):
                        if info["inspect"]["results"][mode].get("errors"):
                            mode_has_error.append(mode)
                        else:
                            mode_ok.append(mode)
                    for mode in mode_ok:
                        info["inspect"]["results"].pop(mode)
                    for mode in mode_has_error:
                        keys = list(info["inspect"]["results"][mode].keys())
                        # remove all except errors
                        for k in keys:
                            if k != "errors":
                                info["inspect"]["results"][mode].pop(k)

                mini["inspect"]["sources"][job] = info

        if src.get("locked"):
            mini["locked"] = src["locked"]
        mini["count"] = count

        return mini

    def get_sources(self, id=None, debug=False, detailed=False):
        dm = self.dump_manager
        um = self.upload_manager
        dpm = self.data_plugin_manager
        ids = set()
        if id and id in dm.register:
            ids.add(id)
        elif id and id in um.register:
            ids.add(id)
        elif id and id in dpm.register:
            ids.add(id)
        else:
            # either no id passed, or doesn't exist
            if id and not len(ids):
                raise ValueError("Source %s doesn't exist" % repr(id))
            ids = set(dm.register)
            ids.update(um.register)
            ids.update(dpm.register)
        sources = {}
        bydsrcs = {}
        byusrcs = {}
        bydpsrcs = {}
        plugins = get_data_plugin().find()
        [bydsrcs.setdefault(src["_id"], src) for src in dm.source_info() if dm]
        [byusrcs.setdefault(src["_id"], src) for src in um.source_info() if um]
        [bydpsrcs.setdefault(src["_id"], src) for src in plugins]
        for _id in ids:
            # start with dumper info
            if dm:
                src = bydsrcs.get(_id)
                if src:
                    if debug:
                        sources[src["name"]] = src
                    else:
                        sources[src["name"]] = self.sumup_source(src, detailed)
            # complete with uploader info
            if um:
                src = byusrcs.get(_id)
                if src:
                    # collection-only source don't have dumpers and only exist in
                    # the uploader manager
                    if not src["_id"] in sources:
                        sources[src["_id"]] = self.sumup_source(src, detailed)
                    if src.get("upload"):
                        for subname in src["upload"].get("jobs", {}):
                            try:
                                sources[src["name"]].setdefault(
                                    "upload",
                                    {"sources": {}})["sources"].setdefault(
                                        subname, {})
                                sources[src["name"]]["upload"]["sources"][
                                    subname]["uploader"] = src["upload"][
                                        "jobs"][subname].get("uploader")
                            except Exception as e:
                                logging.error("Source is invalid: %s\n%s" %
                                              (e, pformat(src)))
            # deal with plugin info if any
            if dpm:
                src = bydpsrcs.get(_id)
                if src:
                    assert len(
                        dpm[_id]
                    ) == 1, "Expected only one uploader, got: %s" % dpm[_id]
                    klass = dpm[_id][0]
                    src.pop("_id")
                    if hasattr(klass, "data_plugin_error"):
                        src["error"] = klass.data_plugin_error
                    sources.setdefault(_id, {"data_plugin": {}})
                    if src.get("download", {}).get("err"):
                        src["download"]["error"] = src["download"].pop("err")
                    sources[_id]["data_plugin"] = src
                    sources[_id]["_id"] = _id
                    sources[_id]["name"] = _id
        if id:
            src = list(sources.values()).pop()
            # enrich with metadata (uploader > dumper)
            ks = []
            if dm:
                try:
                    ks.extend(dm.register[id])
                except KeyError:
                    pass
            if um:
                try:
                    ks.extend(um.register[id])
                except KeyError:
                    pass
            for upk in ks:
                # name either from uploader or dumper
                name = getattr(upk, "name", None) or upk.SRC_NAME
                if getattr(upk, "__metadata__", {}).get("src_meta"):
                    src.setdefault("__metadata__", {}).setdefault(name, {})
                    src["__metadata__"][name] = upk.__metadata__["src_meta"]
            # simplify as needed (if only one source in metadata, remove source key level,
            # or if licenses are the same amongst sources, keep one copy)
            if len(src.get("__metadata__", {})) == 1:
                src["__metadata__"] = list(src["__metadata__"].values()).pop()
            elif len(src.get("__metadata__", {})) > 1:
                metas = list(src["__metadata__"].values())
                simplified = [metas.pop()]
                same = True
                while metas:
                    m = metas.pop()
                    if m not in simplified:
                        same = False
                        break
                if same:
                    # we consume all of them, ie. they're all equals
                    src["__metadata__"] = list(
                        src["__metadata__"].values()).pop()
                else:
                    # convert to a list of dict (so it's easier to detect if one or more
                    # licenses just by checking if type is dict (one) or array (more))
                    metas = src.pop("__metadata__")
                    src["__metadata__"] = []
                    for m in metas:
                        src["__metadata__"].append({m: metas[m]})
            return src
        else:
            return list(sources.values())

    def get_source(self, name, debug=False):
        return self.get_sources(id=name, debug=debug, detailed=True)

    def save_mapping(self, name, mapping=None, dest="master", mode="mapping"):
        logging.debug("Saving mapping for source '%s' destination='%s':\n%s" %
                      (name, dest, pformat(mapping)))
        # either given a fully qualified source or just sub-source
        try:
            subsrc = name.split(".")[1]
        except IndexError:
            subsrc = name
        if dest == "master":
            m = self.src_master.find_one({"_id": subsrc}) or {"_id": subsrc}
            m["mapping"] = mapping
            self.src_master.save(m)
        elif dest == "inspect":
            m = self.src_dump.find_one({"_id": name})
            try:
                m["inspect"]["jobs"][subsrc]["inspect"]["results"][
                    mode] = mapping
                self.src_dump.save(m)
            except KeyError as e:
                raise ValueError(
                    "Can't save mapping, document doesn't contain expected inspection data"
                    % e)
        else:
            raise ValueError("Unknow saving destination: %s" % repr(dest))

    def reset(self, name, key="upload", subkey=None):
        """
        Reset, ie. delete, internal data (src_dump document) for given source name, key subkey.
        This method is useful to clean outdated information in Hub's internal database.

        Ex: key=upload, name=mysource, subkey=mysubsource, will delete entry in corresponding
            src_dump doc (_id=mysource), under key "upload", for sub-source named "mysubsource"

        "key" can be either 'download', 'upload' or 'inspect'. Because there's no such notion of subkey for
        dumpers (ie. 'download', subkey is optional.
        """
        doc = self.src_dump.find_one({"_id": name})
        if not doc:
            raise ValueError("No such datasource named '%s'" % name)
        try:
            # nested
            if key in ["upload", "inspect"]:
                del doc[key]["jobs"][subkey]
            # not nested
            elif key == "download":
                del doc[key]
            else:
                raise ValueError("key=%s not allowed" % repr(key))
            self.src_dump.save(doc)
        except KeyError as e:
            logging.exception(e)
            raise ValueError(
                "Can't delete information, not found in document: %s" % e)



'''
The MIT License (MIT)

Copyright (c) 2014 Ilya Volkov

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

# when comparing list, should we have the same order ?
# default: yes, so no unordered lists
UNORDERED_LIST = False
# when adjusting list diff, should we use list operations
# like (add,replace,move) ? Default: no, bc otherwise
# we can't patch a doc multiple time
USE_LIST_OPS = False


__all__ = ["make",] 

def _store_index(a, x, v):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        try:
            if a[mid][0] < x: lo = mid+1
            else: hi = mid
        except TypeError:
            hi = mid
    if lo < len(a) and a[lo][0] == x:
        a[lo][1].append(v)
    else:
        a.insert(lo, (x, [v]))

def _take_index(a, x):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        try:
            if a[mid][0] < x: lo = mid+1
            else: hi = mid
        except TypeError:
            hi = mid
    if lo < len(a) and a[lo][0] == x:
        if a[lo][1]:
            return a[lo][1].pop()
    return None

class _compare_info(object):
    def __init__(self):
        self.removed = []
        self.added   = []
        self.__root = root = []
        root[:] = [root, root, None]

    def insert(self, op):
        root = self.__root
        last = root[0]
        last[1] = root[0] = [last, root, op]
        return root[0]

    def remove(self, index):
        link_prev, link_next, _ = index
        link_prev[1] = link_next
        link_next[0] = link_prev
        index[:] = []

    def iter_from(self, start):
        root = self.__root
        curr = start[1]
        while curr is not root:
            yield curr[2]
            curr = curr[1]

    def __iter__(self):
        root = self.__root
        curr = root[1]
        while curr is not root:
            yield curr[2]
            curr = curr[1]

    def execute(self):
        root = self.__root
        curr = root[1]
        while curr is not root:
            if curr[1] is not root:
                op_first, op_second = curr[2], curr[1][2]
                if op_first.key == op_second.key and \
                        op_first.path == op_second.path and \
                        type(op_first) == _op_remove and \
                        type(op_second) == _op_add:
                    yield _op_replace(op_second.path, op_second.key, op_second.value).get()
                    curr = curr[1][1]
                    continue
            yield curr[2].get()
            curr = curr[1]   

class _op_base(object):
    def __init__(self, path, key, value):
        self.path  = path
        self.key   = key
        self.value = value

    def __repr__(self):
        return str(self.get())

class _op_add(_op_base):
    def _on_undo_remove(self, path, key):
        if self.path == path:
            if self.key > key:
                self.key += 1
            else:
                key += 1
        return key

    def _on_undo_add(self, path, key):
        if self.path == path:
            if self.key > key:
                self.key -= 1
            else:
                key += 1
        return key

    def get(self):
        return {'op': 'add', 'path': _path_join(self.path, self.key), 'value': self.value}

class _op_remove(_op_base):
    def _on_undo_remove(self, path, key):
        if self.path == path:
            if self.key >= key:
                self.key += 1
            else:
                key -= 1
        return key

    def _on_undo_add(self, path, key):
        if self.path == path:
            if self.key > key:
                self.key -= 1
            else:
                key -= 1
        return key

    def get(self):
        return {'op': 'remove', 'path': _path_join(self.path, self.key)}

class _op_replace(_op_base):
    def _on_undo_remove(self, path, key):
        return key

    def _on_undo_add(self, path, key):
        return key

    def get(self):
        return {'op': 'replace', 'path': _path_join(self.path, self.key), 'value': self.value}

class _op_move(object):
    def __init__(self, oldpath, oldkey, path, key):
        self.oldpath = oldpath
        self.oldkey  = oldkey
        self.path    = path
        self.key     = key

    def _on_undo_remove(self, path, key):
        if self.oldpath == path:
            if self.oldkey >= key:
                self.oldkey += 1
            else:
                key -= 1
        if self.path == path:
            if self.key > key:
                self.key += 1
            else:
                key += 1
        return key

    def _on_undo_add(self, path, key):
        if self.oldpath == path:
            if self.oldkey > key:
                self.oldkey -= 1
            else:
                key -= 1
        if self.path == path:
            if self.key > key:
                self.key -= 1
            else:
                key += 1
        return key

    def get(self):
        return {'op': 'move', 'path': _path_join(self.path, self.key), 'from': _path_join(self.oldpath, self.oldkey)}
    
    def __repr__(self):
        return str(self.get())

def _path_join(path, key):
    if key != None:
        return path + '/' + str(key).replace('~', '~0').replace('/', '~1')
    return path

def _item_added(path, key, info, item):
    index = _take_index(info.removed, item)
    if index != None:
        op = index[2]
        if type(op.key) == int:
            for v in info.iter_from(index):
                op.key = v._on_undo_remove(op.path, op.key)
        info.remove(index)
        if op.path != path or op.key != key:
            new_op = _op_move(op.path, op.key, path, key)
            info.insert(new_op)
    else:
        new_op = _op_add(path, key, item)
        new_index = info.insert(new_op)
        _store_index(info.added, item, new_index)

def _item_removed(path, key, info, item):
    new_op = _op_remove(path, key, item)
    index = _take_index(info.added, item)
    new_index = info.insert(new_op)
    if index != None:
        op = index[2]
        if type(op.key) == int:
            for v in info.iter_from(index):
                op.key = v._on_undo_add(op.path, op.key)
        info.remove(index)
        if new_op.path != op.path or new_op.key != op.key:
            new_op = _op_move(new_op.path, new_op.key, op.path, op.key)
            new_index[2] = new_op
        else:
            info.remove(new_index)
    else:
        _store_index(info.removed, item, new_index)

def _item_replaced(path, key, info, item):
    info.insert(_op_replace(path, key, item))

def _compare_dicts(path, info, src, dst):
    added_keys = dst.keys() - src.keys()
    removed_keys = src.keys() - dst.keys()
    for key in removed_keys:
        _item_removed(path, str(key), info, src[key])
    for key in added_keys:
        _item_added(path, str(key), info, dst[key])
    for key in src.keys() & dst.keys():
        _compare_values(path, key, info, src[key], dst[key])

def _compare_lists(path, info, src, dst):
    len_src, len_dst = len(src), len(dst)
    if USE_LIST_OPS:
        max_len = max(len_src, len_dst)
        min_len = min(len_src, len_dst)
        for key in range(max_len):
            if key < min_len:
                old, new = src[key], dst[key]
                if old == new:
                    continue
                _item_removed(path, key, info, old)
                _item_added(path, key, info, new)
            elif len_src > len_dst:
                _item_removed(path, len_dst, info, src[key])
            else:
                _item_added(path, key, info, dst[key])
    else:
        if len_src != len_dst or (not UNORDERED_LIST and src != dst):
            _item_replaced(path, None , info, dst)
        else:
            found_diff = False
            # lengths are the same so we just need to compare src against dst
            # (dst against src isn't necessary)
            for e in src:
                if not e in dst:
                    found_diff = True
                    break
            if found_diff:
                _item_replaced(path, None , info, dst)


def _compare_values(path, key, info, src, dst):
    if src == dst:
        return
    elif isinstance(src, dict) and \
            isinstance(dst, dict):
        _compare_dicts(_path_join(path, key), info, src, dst)
    elif isinstance(src, list) and \
            isinstance(dst, list):
        _compare_lists(_path_join(path, key), info, src, dst)
    else:
        _item_replaced(path, key, info, dst)

def make(src, dst, **kwargs):
    info = _compare_info()
    _compare_values('', None, info, src, dst)
    return [op for op in info.execute()]




''' Some utility functions that do document traversal '''
from biothings.utils.common import is_seq

# doc labelled in breadth first way with letters
test_doc = {
    'a': {
        'b': {
            'e': {
                'i': 'i value',
                'j': 'j value'
            },
            'f': 'f value'
        },
        'c': 'c value',
        'd': {
            'g': {
                'k': 'k value',
                'l': 'l value'
            },
            'h': 'h value'
        }
    }
}

class QueueEmptyError(Exception):
    pass


class StackEmptyError(Exception):
    pass


class Stack(object):
    def __init__(self):
        self.stack = []

    def push(self, obj):
        ''' put obj on stack '''
        self.stack.insert(0, obj)

    def pop(self):
        try:
            return self.stack.pop(0)
        except IndexError:
            raise StackEmptyError("Can't pop object from an empty stack")

    def isempty(self):
        return len(self.stack) == 0


class Queue(object):
    def __init__(self):
        self.queue = []

    def push(self, obj):
        ''' put obj on queue '''
        self.queue.append(obj)

    def pop(self):
        ''' get next obj from queue '''
        try:
            return self.queue.pop(0)
        except IndexError:
            # empty queue
            raise QueueEmptyError("Can't pop object from an empty queue")

    def isempty(self):
        return len(self.queue) == 0

def breadth_first_traversal(doc):
    ''' Yield a 2 element tuple for every k, v pair in document items (nodes are visited in breadth first order
        k is itself a tuple of keys annotating the path for this node (v) to root
        v is the node value
    '''
    return _generic_traversal(doc, Queue)

def depth_first_traversal(doc):
    ''' Yield a 2 element tuple for every k, v pair in document items (nodes are visited in depth first order
        k is itself a tuple of keys annotating the path for this node (v) to root
        v is the node value
    '''
    return _generic_traversal(doc, Stack)

def _generic_traversal(doc, structure):
    _struct = structure()

    # push first level
    for (k, v) in doc.items():
        # _struct.push((tuple([k]), v))   # TODO: remove this line
        _struct.push(((k,), v))

    while not _struct.isempty():
        _next = _struct.pop()
        yield _next
        if isinstance(_next[1], dict):
            # push this level
            for (k, v) in _next[1].items():
                _struct.push((tuple(list(_next[0]) + [k]), v))
        elif is_seq(_next[1]):
            # push all elements in a list/tuple
            for o in _next[1]:
                _struct.push((_next[0], o))

def breadth_first_recursive_traversal(doc, path=None):
    ''' doesn't exactly implement breadth first ordering it seems, not sure why... '''
    #TODO fix this...
    path = path or []
    if isinstance(doc, dict):
        for (k, v) in doc.items():
            yield (tuple(list(path) + [k]), v)
        for (k, v) in doc.items():
            yield from breadth_first_recursive_traversal(v, tuple(list(path) + [k]))
    elif is_seq(doc):
        for o in doc:
            yield (tuple(list(path)), o)
        for o in doc:
            yield from breadth_first_recursive_traversal(o, tuple(list(path)))

def depth_first_recursive_traversal(doc, path=None):
    path = path or []
    if isinstance(doc, dict):
        for (k, v) in doc.items():
            _path = tuple(list(path) + [k])
            yield (_path, v)
            yield from depth_first_recursive_traversal(v, _path)
    elif is_seq(doc):
        for o in doc:
            _path = tuple(list(path))
            yield (_path, o)
            yield from depth_first_recursive_traversal(o, _path)



from .dotstring import list_length
from .dotstring import remove_key


class ExcludeFieldsById(object):
    """
    This class provides a framework to exclude fields for certain
    identifiers. Up to three arguments are passed to this class, an
    identifier list, a list of fields to remove, and minimum list
    size.  The identifier list is a list of document identifiers to act
    on.  The list of fields are fields that will be removed; they are
    specified using a dotstring notation.  The minimum list size is
    the minimum number of elements that should be in a list in order
    for it to be removed.  The 'drugbank', 'chebi', and 'ndc' data
    sources were manually tested with this class.
    """

    def __init__(self, exclusion_ids, field_lst, min_list_size=1000):
        """
        Fields to truncate are specified by field_lst.  The
        dot-notation is accepted.
        """
        self.exclusion_ids = exclusion_ids
        self.field_lst = field_lst
        self.min_list_size = min_list_size

    def __call__(self, f):
        """
        Truncate specified fields on documents on call.
        :param f: function to apply to, this function should return documents
        :return:
        """
        def wrapped_f(*args):
            input_docs = f(*args)
            for doc in input_docs:
                if doc['_id'] in self.exclusion_ids:
                    for field in self.field_lst:
                        # min_list_size check
                        if list_length(doc, field) > self.min_list_size:
                            remove_key(doc, field)
                yield doc
        return wrapped_f



'''
NOTE: biothings.utils.shelve is deprecated!
'''
raise Exception("Don't use me, no concurrency supported...")

import os
import shelve

from biothings import config
from biothings.utils.hub_db import IConnection
from biothings.utils.dotfield import parse_dot_fields
from biothings.utils.dataload import update_dict_recur


def get_src_conn():
    return Database(config.DATA_SRC_DATABASE)

def get_src_dump():
    db = Database(config.DATA_SRC_DATABASE)
    return db[config.DATA_SRC_DUMP_COLLECTION]

def get_src_build():
    db = Database(config.DATA_SRC_DATABASE)
    return db[config.DATA_SRC_BUILD_COLLECTION]

def get_src_build_config():
    db = Database(config.DATA_SRC_DATABASE)
    return db[config.DATA_SRC_BUILD_CONFIG_COLLECTION]

# shelve won't allow opening db file in 2 different threads
# so we want to make sure to use singleton pattern here
_db_register = {}
class Database(Connection):
    def __new__(cls, dbname):
        global _db_register
        if not dbname in _db_register:
            obj = Connection.__new__(cls)
            obj.name = dbname
            _db_register[dbname] = obj
        return _db_register[dbname]

    def __init__(self,dbname):
        super(Database,self).__init__()
        self.name = dbname
        self.cols = {}
        self.dbfolder = os.path.join(config.SHELVE_DB_FOLDER,dbname)

    def collection_names(self):
        # for shelve, it's just the db file in db folder
        return os.listdir(self.dbfolder)

    def create_collection(self,colname):
        return self[colname]

    def __getitem__(self, colname):
        if not colname in self.cols:
            self.cols[colname] = Collection(self.name,colname,self.dbfolder)
        return self.cols[colname]

    def __repr__(self):
        return "<%s at %s, %s>" % (self.__class__.__name__,hex(id(self)),self.name)


class Collection(object):

    def __init__(self, dbname, colname, dbfolder):
        if not os.path.exists(dbfolder):
            os.makedirs(dbfolder)
        self.colname = colname
        self.name = dbname
        self.shelf = shelve.open(os.path.join(dbfolder,self.colname))

    @property
    def name(self):
        return self.colname

    @property
    def database(self):
        return Database(self.name)

    def find_one(self,*args,**kwargs):
        if args and len(args) == 1 and type(args[0]) == dict:
            if len(args[0]) == 1 and "_id" in args[0]:
                return self.shelf.get(str(args[0]["_id"])) # keys in shelve are str only...
            else:
                return self.find(*args,find_one=True)
        elif args or kwargs:
            raise NotImplementedError("find(): %s %s" % (repr(args),repr(kwargs)))

    def find(self,*args,**kwargs):
        results = []
        if args and len(args) == 1 and type(args[0]) == dict:
            # it's key/value search, let's iterate
            for doc in self.shelf.values():
                found = False
                for k,v in args[0].items():
                    if k in doc:
                        if doc[k] == v:
                            found = True
                    else:
                        found = False
                        break
                if found:
                    if "find_one" in kwargs:
                        return doc
                    else:
                        results.append(doc)
            return results
        elif args:
            raise NotImplementedError("find(): %s %s" % (repr(args),repr(kwargs)))
        else:
            return (v for v in self.shelf.values())

    def insert_one(self,doc):
        assert "_id" in doc
        sid = str(doc["_id"])
        if sid in self.shelf:
            raise Exception("Duplicated key error '%s'" % sid)
        self.shelf[sid] = doc

    def save(self,doc):
        sid = str(doc["_id"])
        self.shelf[sid] = doc

    def update_one(self,query,what):
        assert len(what) == 1 and "$set" in what
        doc = self.find_one(query)
        if not doc:
            raise Exception("No result for query %s" % repr(query))
        what = parse_dot_fields(what)
        doc = update_dict_recur(doc,what)
        self.save(doc)

    def count(self):
        return len(self.shelf)

    def __getitem__(self, _id):
        return self.cols[colname]



import os
import time
import datetime
import logging
import asyncio
from functools import partial

from .slack import slack_msg

def setup_default_log(default_logger_name, log_folder, level=logging.DEBUG):
    # this will affect any logging calls
    logging.basicConfig(level=level)
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)
    logfile = os.path.join(log_folder, '%s_%s_hub.log' % (default_logger_name, time.strftime("%Y%m%d", datetime.datetime.now().timetuple())))
    fh = logging.FileHandler(logfile)
    fh.setFormatter(logging.Formatter('%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s', datefmt="%H:%M:%S"))
    fh.name = "logfile"
    logger = logging.getLogger(default_logger_name)
    logger.setLevel(level)
    if fh.name not in [h.name for h in logger.handlers]:
        logger.addHandler(fh)
    return logger


def get_logger(logger_name, log_folder=None, handlers=("console", "file", "slack"), timestamp="%Y%m%d"):
    """
    Configure a logger object from logger_name and return (logger, logfile)
    """
    from biothings import config as btconfig
    if not log_folder:
        log_folder = btconfig.LOG_FOLDER
    # this will affect any logging calls
    if not os.path.exists(log_folder):
        os.makedirs(log_folder)
    if timestamp:
        logfile = os.path.join(log_folder, '%s_%s.log' % (logger_name, time.strftime(timestamp, datetime.datetime.now().timetuple())))
    else:
        logfile = os.path.join(log_folder, '%s.log' % logger_name)
    fmt = logging.Formatter('%(asctime)s [%(process)d:%(threadName)s] - %(name)s - %(levelname)s -- %(message)s', datefmt="%H:%M:%S")
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    if "file" in handlers:
        fh = logging.FileHandler(logfile)
        fh.setFormatter(fmt)
        fh.name = "logfile"
        if fh.name not in [h.name for h in logger.handlers]:
            logger.addHandler(fh)

    if "hipchat" in handlers:
        raise DeprecationWarning("Hipchat is dead...")

    if "slack" in handlers and getattr(btconfig, "SLACK_WEBHOOK", None):
        nh = SlackHandler(btconfig.SLACK_WEBHOOK)
        nh.setFormatter(fmt)
        nh.name = "slack"
        if nh.name not in [h.name for h in logger.handlers]:
            logger.addHandler(nh)

    return (logger, logfile)


class SlackHandler(logging.StreamHandler):

    colors = {
        logging.DEBUG: "#a1a1a1",
        logging.INFO: "good",
        logging.WARNING: "warning",
        logging.ERROR: "danger",
        logging.CRITICAL: "#7b0099"
    }

    def __init__(self, webhook):
        super(SlackHandler, self).__init__()
        self.webhook = webhook

    def emit(self, record):
        @asyncio.coroutine
        def aioemit():
            fut = yield from loop.run_in_executor(None,
                                                  partial(slack_msg, self.webhook, msg, color=color))
            return fut
        if record.__dict__.get("notify"):
            try:
                loop = asyncio.get_event_loop()
            except RuntimeError:
                # if the logger is running in a thread, there's no asyncio loop there
                # (we usually take it from job_manager.loop, but it's not accessible there)
                # so we use another loop
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            msg = self.format(record)
            color = self.__class__.colors.get(record.levelno, self.__class__.colors[logging.DEBUG])
            fut = aioemit()
            asyncio.ensure_future(fut)


class EventRecorder(logging.StreamHandler):

    def __init__(self, *args, **kwargs):
        super(EventRecorder, self).__init__(*args, **kwargs)
        from biothings.utils.hub_db import get_event
        self.eventcol = get_event()

    def emit(self, record):
        @asyncio.coroutine
        def aioemit(msg):
            def recorded(f):
                res = f.result()
            fut = loop.run_in_executor(
                None,
                partial(self.eventcol.save, msg)
            )
            fut.add_done_callback(recorded)
            yield from fut
            return fut
        if record.__dict__.get("notify") or record.__dict__.get("event"):
            try:
                loop = asyncio.get_event_loop()
                msg = {
                    "_id": record.created,
                    "asctime": record.asctime,
                    "msg": record.message,
                    "level": record.levelname,
                    "name": record.name,
                    "pid": record.process,
                    "pname": record.processName
                }
                fut = aioemit(msg)
                asyncio.ensure_future(fut)
            except Exception as e:
                logging.error("Couldn't record event: %s", e)


class WSLogHandler(logging.StreamHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def __init__(self, listener):
        super().__init__()
        self.listener = listener
        self.count = 0

    def payload(self, record):
        msg = self.format(record)
        return {
            "_id": self.count,
            "op": "log",
            "msg": msg,
            "logger": record.name,
            "level": record.levelname,
            "ts": datetime.datetime.now().isoformat()
        }

    def emit(self, record):
        self.count += 1
        self.listener.read(self.payload(record))

class WSShellHandler(WSLogHandler):
    """
    when listener is a bt.hub.api.handlers.ws.LogListener instance,
    log statements are propagated through existing websocket
    """

    def payload(self, record):
        types = {
            ShellLogger.INPUT: "input",
            ShellLogger.OUTPUT: "output"
        }
        return {
            "_id": self.count,
            "op": "shell",
            "cmd": record.msg,
            "type": types.get(record.levelno, "unknown"),
            "ts": datetime.datetime.now().isoformat()
        }


class ShellLogger(logging.Logger):
    """
    Custom "levels" for input going to the shell
    and output coming from it (just for naming)
    """

    OUTPUT = 1000
    INPUT = 1001

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.manager.loggerDict[self.name] = self

    def input(self, msg, *args, **kwargs):
        self._log(self.__class__.INPUT, msg, args, **kwargs)

    def output(self, msg, *args, **kwargs):
        self._log(self.__class__.OUTPUT, msg, args, **kwargs)
